import types
from math import factorial

import numpy as np
import pytest
from numpy.typing import ArrayLike

from fastfem.elements.linear_simplex2d import LinearSimplex2D
from fastfem.fields.field import Field


def meshquad(X, Y, F):
    """
    In case this is ever needed, this is a function to compute the integral on a 2D mesh.
    Integrals are estimated by estimating integrals on triangles, which are defined by
    indices ([i,j] - [i+1,j] - [i,j+1]) ([i+1,j+1] - [i+1,j] - [i,j+1])
    """

    def cross_mag(XA, YA, XB, YB):
        return np.abs(XA * YB - XB * YA)

    F00 = F[:-1, :-1]
    F01 = F[:-1, 1:]
    F10 = F[1:, :-1]
    F11 = F[1:, 1:]
    X00 = X[:-1, :-1]
    X01 = X[:-1, 1:]
    X10 = X[1:, :-1]
    X11 = X[1:, 1:]
    Y00 = Y[:-1, :-1]
    Y01 = Y[:-1, 1:]
    Y10 = Y[1:, :-1]
    Y11 = Y[1:, 1:]
    axes = np.arange(X.ndim)
    return np.sum(
        cross_mag(X10 - X00, Y10 - Y00, X01 - X00, Y01 - Y00) * (F00 + F01 + F10) / 6,
        axes,
    ) + np.sum(
        cross_mag(X10 - X11, Y10 - Y11, X01 - X11, Y01 - Y11) * (F11 + F01 + F10) / 6,
        axes,
    )


def _analytic_integrate_reftri(powx: int, powy: int) -> float:
    """Integrates x**powx * y**powy on the reference triangle
    [0,0] - [1,0] - [0,1]

    Args:
        powx (int): exponent for x
        powy (int): exponent for y

    Returns:
        float: the resultant integral
    """
    #   int_0^1 x**powx int_0^{1-x}  y**powy   dy dx
    return factorial(powx) * factorial(powy) / factorial(powx + powy + 2)


@pytest.fixture(scope="module")
def element():
    return LinearSimplex2D()


@pytest.fixture(scope="module")
def transformed_element(element, transformation):
    return (
        element,
        Field(
            element.basis_shape(),
            (2,),
            transformation(element.reference_element_position_field().coefficients),
        ),
        transformation,
    )


@pytest.fixture(scope="session")
def triangle_quadrature():
    reference_rules = dict()

    class TriangleQuadratureRuleBuilder:
        """A triangular quadrature rule tuned to a specific
        degree of exactness.
        """

        def __call__(
            self, exactness: int, tribounds: ArrayLike = [[0, 0], [1, 0], [0, 1]]
        ):
            if not isinstance(tribounds, np.ndarray):
                tribounds = np.array(tribounds)

            if exactness not in reference_rules:
                # num polys: count (a,b) for which a+b <= exactness
                # for simplicity of solving for weights, we use a
                # 2d newton-coates variant, where our "equally-spaced"
                # nodes are given as
                def line(y, n):
                    pts = np.empty((n, 2))
                    pts[:, 0] = np.linspace(0, 1 - y, n)
                    pts[:, 1] = y
                    return pts

                meshpts = np.concatenate(
                    [
                        line(1 - n / (exactness + 1), n + 1)
                        for n in range(exactness + 1)
                    ],
                    axis=0,
                )
                integrals = np.empty(meshpts.shape[0])
                coef_evals = np.empty((meshpts.shape[0], meshpts.shape[0]))

                iterm: int = 0
                for powx in range(exactness + 1):
                    for powy in range(exactness - powx + 1):
                        coef_evals[iterm, :] = (
                            meshpts[:, 0] ** powx * meshpts[:, 1] ** powy
                        )
                        integrals[iterm] = _analytic_integrate_reftri(powx, powy)
                        iterm += 1

                # CE_{ij} w_j = I_i
                # where CE_{ij} = f_i(x_j),  I_i = int(f_i)
                weights = np.linalg.solve(coef_evals, integrals)
                reference_rules[exactness] = types.SimpleNamespace(
                    weights=weights, knots=meshpts
                )

            # we have the reference element values. we want quadratures on tribounds
            affine_shift = tribounds[0, :]
            affine_linT = tribounds[1:, :] - tribounds[0, :]

            weights = reference_rules[exactness].weights * abs(
                np.linalg.det(affine_linT)
            )
            knots = affine_shift + np.einsum(
                "ij,...i->...j", affine_linT, reference_rules[exactness].knots
            )

            return types.SimpleNamespace(
                weights=weights, knots=knots, exactness=exactness, tribounds=tribounds
            )

    return TriangleQuadratureRuleBuilder()


@pytest.mark.skip
def test_integrate_reftri(triangle_quadrature):
    # this is used to do midpoint rule on triangles
    def cross_mag(A, B):
        return np.abs(A[..., 0] * B[..., 1] - B[..., 0] * A[..., 1])

    # 10 subdivs -> ~3 mil points?
    n_subdivisions = 10
    reftri_mesh = np.array([[[0, 0], [1, 0], [0, 1]]])
    for _ in range(n_subdivisions):
        mid01 = 0.5 * (reftri_mesh[:, 0, :] + reftri_mesh[:, 1, :])
        mid12 = 0.5 * (reftri_mesh[:, 2, :] + reftri_mesh[:, 1, :])
        mid20 = 0.5 * (reftri_mesh[:, 0, :] + reftri_mesh[:, 2, :])
        numtris = reftri_mesh.shape[0]
        mesh_next = np.empty((numtris * 4, 3, 2))

        mesh_next[:numtris, 0, :] = reftri_mesh[:, 0, :]
        mesh_next[:numtris, 1, :] = mid01
        mesh_next[:numtris, 2, :] = mid20

        mesh_next[numtris : (2 * numtris), 0, :] = mid01
        mesh_next[numtris : (2 * numtris), 1, :] = reftri_mesh[:, 1, :]
        mesh_next[numtris : (2 * numtris), 2, :] = mid12

        mesh_next[(2 * numtris) : (3 * numtris), 0, :] = mid20
        mesh_next[(2 * numtris) : (3 * numtris), 1, :] = mid12
        mesh_next[(2 * numtris) : (3 * numtris), 2, :] = reftri_mesh[:, 2, :]

        mesh_next[(3 * numtris) :, 0, :] = mid12
        mesh_next[(3 * numtris) :, 1, :] = mid20
        mesh_next[(3 * numtris) :, 2, :] = mid01
        reftri_mesh = mesh_next

    midpoints = np.mean(reftri_mesh, axis=1)

    for powx in range(4):
        for powy in range(4):
            res = _analytic_integrate_reftri(powx, powy)
            analytic = np.sum(
                # f(x,y) at midpoints
                (midpoints[:, 0] ** powx * midpoints[:, 1] ** powy)
                # areas
                * 0.5
                * cross_mag(
                    reftri_mesh[:, 1, :] - reftri_mesh[:, 0, :],
                    reftri_mesh[:, 2, :] - reftri_mesh[:, 0, :],
                )
            )
            assert res == pytest.approx(
                analytic, rel=1e-5
            ), f"Failed _analytic_integrate_reftri({powx},{powy})"
            for exactdeg in range(powx + powy, 10):
                quad_rule = triangle_quadrature(exactdeg)
                quad_result = np.sum(
                    quad_rule.weights
                    * (quad_rule.knots[:, 0] ** powx * quad_rule.knots[:, 1] ** powy)
                )
                assert quad_result == pytest.approx(
                    res, rel=1e-9
                ), f"Failed exactness for x**{powx} * y**{powy} for DOE {exactdeg}."


def test_interpolate_field(element):
    field = element.basis_fields()
    constfield = Field(tuple(), tuple(), 3)
    for x, y in [(0, 0), (1, 0), (0, 1), (0.5, 0.5), (0.25, 0.25)]:
        field_outs = np.array([1 - x - y, x, y])
        np.testing.assert_allclose(
            element.interpolate_field(field, x, y).coefficients, field_outs
        )

        assert element.interpolate_field(
            constfield, x, y
        ).coefficients == pytest.approx(constfield.coefficients)


def test_integrate_field(transformed_element, triangle_quadrature):
    element, pts, transform = transformed_element
    quad = triangle_quadrature(3, tribounds=pts.coefficients)
    ref_quad = triangle_quadrature(3)

    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(), tuple(), field.coefficients[:, np.newaxis, :]
    )
    result = element.integrate_field(pts, field, jac_scale).coefficients

    field_quad = element.interpolate_field(
        field.stack[np.newaxis, :],
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients
    jac_quad = element.interpolate_field(
        jac_scale,
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients

    quad_result = np.sum(quad.weights[:] * field_quad * jac_quad, axis=-1)

    np.testing.assert_allclose(result, quad_result, atol=1e-10)


def test_integrate_basis_times_field(transformed_element, triangle_quadrature):
    element, pts, transform = transformed_element
    quad = triangle_quadrature(4, tribounds=pts.coefficients)
    ref_quad = triangle_quadrature(4)

    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(), tuple(), field.coefficients[:, np.newaxis, :]
    )
    result = element.integrate_basis_times_field(pts, field, None, jac_scale)

    field_quad = element.interpolate_field(
        field,
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients
    jac_quad = element.interpolate_field(
        jac_scale,
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients
    basis_field = Field(
        element.basis_shape(), tuple(), field.coefficients[:, np.newaxis, np.newaxis, :]
    )
    basis_quad = element.interpolate_field(
        basis_field,
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients

    quad_result = np.sum(
        quad.weights * field_quad * jac_quad * basis_quad,
        axis=-1,
    )

    np.testing.assert_allclose(result.coefficients, quad_result, atol=1e-10)


def test_integrate_grad_basis_dot_field(transformed_element, triangle_quadrature):
    element, pts, transform = transformed_element
    quad = triangle_quadrature(4, tribounds=pts.coefficients)
    ref_quad = triangle_quadrature(4)

    field = element.basis_fields()
    jac_scale = Field(
        element.basis_shape(), tuple(), field.coefficients[:, np.newaxis, :]
    )
    basis_field = Field(
        element.basis_shape(), tuple(), field.coefficients[:, np.newaxis, np.newaxis, :]
    )
    grad_basis = element.compute_field_gradient(basis_field, pts)
    basis_quad = element.interpolate_field(
        grad_basis,
        ref_quad.knots[:, 0],
        ref_quad.knots[:, 1],
    ).coefficients
    for eind in range(2):
        efield = Field(
            field.basis_shape,
            (2,),
            field.coefficients[..., np.newaxis] * np.eye(2)[:, eind],
        )

        result = element.integrate_grad_basis_dot_field(pts, efield, None, jac_scale)

        field_quad = element.interpolate_field(
            efield,
            ref_quad.knots[:, 0],
            ref_quad.knots[:, 1],
        ).coefficients
        jac_quad = element.interpolate_field(
            jac_scale,
            ref_quad.knots[:, 0],
            ref_quad.knots[:, 1],
        ).coefficients

        quad_result = np.moveaxis(
            np.sum(
                quad.weights[:]
                * jac_quad
                * np.einsum("...i,...i->...", field_quad, basis_quad),
                axis=-1,
            ),
            0,
            -1,
        )  # shift basis to right of stack

        np.testing.assert_allclose(result.coefficients, quad_result, atol=1e-10)


def test_field_gradient_const(element):
    constfield = Field(tuple(), tuple(), 3)
    np.testing.assert_allclose(
        element._compute_field_gradient(constfield).coefficients, 0, atol=1e-10
    )



import collections.abc as colltypes
import typing

import numpy as np
import numpy.typing
from numpy.typing import ArrayLike, NDArray

import fastfem.elements.element2d as element
import fastfem.fields.numpy_similes as fnp
from fastfem.fields.field import Field, FieldAxisIndex, ShapeComponent


class DeformationGradient2DBadnessException(Exception):
    """An exception called when the deformation gradient is too poor for invertibility.

    Args:
        val (float): a badness parameter det(F) / (`char_x` * `char_y`) for
            characteristic lengths `char_x` and `char_y`. This may be in reference
            to the size of the element, so that scaling the element yields the same
            badness parameter.
        x (float): the x position in local (reference) coordinates of the determinant.
        y (float): the y position in local (reference) coordinates of the determinant.
    """

    def __init__(self, val: float, x: float, y: float):
        super().__init__(
            "Element has too poor of a shape!\n"
            f"   def_grad badness = {val:e} at local coordinates"
            f"({x:.6f},{y:.6f})"
        )

        self.x = x
        self.y = y
        self.val = val


def _build_GLL(n: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Builds the items necessary for GLL quadrature of degree `n` (`n+1`-node).

    Args:
        n (int): The degree of the quadrature. This results in `n+1` nodes for the
            quadrature.


    Returns:
        tuple[np.ndarray,np.ndarray,np.ndarray]: A tuple `(x,w,L)` for knots
            (node positions) `x`, weights `w`, and Lagrange interpolation polynomials
            `L` associated with the quadrature. A quadrature for a vectorized function f
            is computed as `sum(f(x) * w)`, while the $i^{th}$ Lagrange polynomial is
            $L_i(x) = \\sum_{k=0}^{n} L[i,k] x^k$.
    """
    if n == 1:
        return np.array((-1, 1)), np.array((1, 1)), np.array([[0.5, -0.5], [0.5, 0.5]])

    leg_x = (
        np.polynomial.Polynomial(np.polynomial.legendre.leg2poly([0] * n + [1]))
        .deriv()
        .roots()
    )
    np1 = n + 1
    x = np.array([-1, *leg_x, 1])

    # initialize L to 1; operate in quad precision at derivation
    L = np.zeros((np1, np1), dtype=np.longdouble)
    L[:, 0] = 1

    tmp = np.empty((n, n), dtype=np.longdouble)
    for i, xi in enumerate(x):
        mask = np.arange(np1) != i

        # what we multiply (x-xi) to
        tmp[:, :] = L[mask, :-1]
        tmp[:, :] /= x[mask, np.newaxis] - xi

        # add tmp*(-xi)
        L[mask, :-1] = tmp * -xi
        # add tmp*x
        L[mask, 1:] += tmp

    # return to double precision
    L = L.astype(np.float64)

    # compute weights by integrating L
    endpoint_pows = (
        np.array([-1, 1])[:, np.newaxis] ** np.arange(1, np1 + 1)[np.newaxis, :]
    )
    endpoints_integ = np.einsum(
        "ik,k,jk->ij", L, 1 / np.arange(1, np1 + 1), endpoint_pows
    )

    return x, endpoints_integ[:, 1] - endpoints_integ[:, 0], L


class SpectralElement2D(element.Element2D):
    """A spectral element in 2 dimensions of order N, leading to (N+1)^2 nodes.
    GLL quadrature is used to diagonalize the mass matrix.
    """

    def __init__(self, degree: int):
        super().__init__()
        self.degree = degree  # poly degree (#nodes - 1)
        self.num_nodes = degree + 1

        self.knots, self.weights, self._lagrange_polys = _build_GLL(degree)
        # store derivatives of L_i as needed, so they need only be computed once
        self._lagrange_derivs = {}
        self._lagrange_derivs[0] = self._lagrange_polys

        self._lagrange_at_knots = {}

    @typing.override
    def basis_shape(self) -> tuple[int, ...]:
        """Returns a tuple representing the shape of the array corresponding to the
        basis coefficients. A scalar field `f`, given as an array is expected to have
        shape `f.shape == element.basis_shape()`
        """
        return (self.num_nodes, self.num_nodes)

    @typing.override
    def reference_element_position_field(self) -> Field:
        """
        The position field of the reference (un-transformed) element. This is a vector
        field.

        Returns:
            Field: An array of shape `(*element.basis_shape(), 2)`
        """
        out = np.empty((self.num_nodes, self.num_nodes, 2))
        out[:, :, 0] = self.knots[:, np.newaxis]
        out[:, :, 1] = self.knots[np.newaxis, :]
        return self.Field(out, False, (2,))

    def lagrange_poly1D(self, deriv_order: int = 0) -> NDArray:
        """
        Returns the polynomial coefficients `P[i,k]`, where
        $\\frac{d^{r}}{dx^{r}} L_{i}(x) = \\sum_k P[i,k] x^k$ with $r$ as `deriv_order`.

        deriv_order (default 0)

        Args:
            deriv_order (int, optional): The order $r$ of the derivative. This is
                expected to be an integer between 0 (inclusive) and degree+1 (exclusive),
                but this check is not done. Defaults to 0.

        Returns:
            np.ndarray: The coefficient array `P[i,k]` which is of shape
            `(degree + 1, degree + 1 - deriv_order)`
        """
        if deriv_order in self._lagrange_derivs:
            return self._lagrange_derivs[deriv_order]

        # inefficient partial partition calc, but only done once, so...
        coefs = self._lagrange_polys
        for _ in range(deriv_order):
            coefs = coefs[:, 1:] * np.arange(1, coefs.shape[1])

        self._lagrange_derivs[deriv_order] = coefs
        return coefs

    @typing.override
    def _interpolate_field(
        self,
        field: Field,
        X: NDArray,
        Y: NDArray,
    ) -> typing.Any:
        """Evaluates field at (X,Y) in reference coordinates.
        The result is an array of values `field(X,Y)`.

        `field` is of shape `(*basis_shape,...,*fieldshape)`, where the internal
        ellipses `...` broadcasts with X and Y
        using numpy's rules for broadcasting.

        X and Y must have compatible shape, broadcasting to pointshape.

        Args:
            field (Field): an array of shape (degree+1,degree+1,*fieldshape)
                representing the field to be interpolated.
            X (NDArray): x values (in reference coordinates).
            Y (NDArray): y values (in reference coordinates).
            fieldshape (tuple[int,...], optional): the shape of `field` at each point.
                Defaults to tuple() for a scalar field.

        Returns:
            typing.Any: The interpolated values `field(X,Y)`
        """
        # F^{i,j} L_{i,j}(X,Y)
        # lagrange_polys[i,k] : component c in term cx^k of poly i
        return fnp.einsum(
            ShapeComponent.BASIS,
            "ij,ia,...a,jb,...b->...",
            field,
            self._lagrange_polys,
            np.expand_dims(X, -1) ** np.arange(self.num_nodes),
            self._lagrange_polys,
            np.expand_dims(Y, -1) ** np.arange(self.num_nodes),
        )

    # @warnings.deprecated("This can be done (probably cleaner) with JAX.")
    def locate_point(
        self,
        pos_field: np.ndarray,
        posx: float,
        posy: float,
        tol: float = 1e-8,
        dmin: float = 1e-7,
        max_iters: int = 1000,
        def_grad_badness_tol: float = 1e-4,
        ignore_out_of_bounds: bool = False,
        char_x: float | None = None,
        char_y: float | None = None,
    ) -> tuple[np.ndarray, bool]:
        """
        Attempts to find the local coordinates corresponding to the
        given global coordinates (posx,posy). Returns (local_pt,success).
        If a point is found, the returned value is ((x,y),True),
        with local coordinates (x,y).
        Otherwise, there are two cases:
          - ((x,y),False) is returned when descent leads out of the
            domain. A step is forced to stay within the local domain
            (max(|x|,|y|) <= 1), but if a constrained minimum is found on an
            edge with a descent direction pointing outwards, that point
            is returned. If ignore_out_of_bounds is true, the interior checking
            does not occur.
          - ((x,y),False) is returned if a local minimum is found inside the
            domain, but the local->global transformation is too far. This
            should technically not occur if the the deformation gradient stays
            nonsingular, but in the case of ignore_out_of_bounds == True, the
            everywhere-invertibility may not hold.

        The initial guess is chosen as the closest node, and a Newton-Raphson
        step is used along-side a descent algorithm. tol is the stopping parameter
        triggering when the error function (ex^2 + ey^2)/2 < tol in global
        coordinates. dmin is the threshold for when a directional derivative
        is considered zero, providing the second stopping criterion.

        def_grad_badness_tol parameterizes how poorly shaped the element is locally.
        If the coordinate vectors line up close enough,
        or one of the vectors gets too small,
        we can catch that with the expression
        (abs(det(def_grad)) < def_grad_badness_tol*char_x*char_y ),
        and raise an exception.
        Here, char_x and char_y are characteristic lengths of the element,
        and are calculated from pos_field when not defined.

        Args:
            pos_field (np.ndarray): an array representing the positions of the
                    element nodes. This is of the shape `(*basis_shape,2)`.
            posx (float): the x coordinate in global coordinates to find.
            posy (float): the y coordinate in global coordinates to find.
            tol (float, optional): Tolerance of the error function. Defaults to 1e-8.
            dmin (float, optional): The largest size of the gradient for a point to be
                    considered a local minimum. If the descent direction r dotted with
                    the gradient of the error function is less than dmin, then the local
                    minimum condition is met (If the gradient points out of the domain
                    then the dot product may be zero). Defaults to 1e-7.
            max_iters (int, optional): The maximum number of iterations taken.
                    Terminates afterwards, returning ((x,y),False).
            def_grad_badness_tol (float, optional): The minimum allowable badness
                    parameter, after which an error is raised. Defaults to 1e-4.
            ignore_out_of_bounds (bool, optional): Whether or not descent directions can
                    point outside of the domain. If False, then locate_point stays
                    within the element. Defaults to False.
            char_x (float | None, optional): characteristic x-length of the element.
                    When not set, a characteristic value is computed from pos_field,
                    set to approximately the largest length curve of the position field
                    along constant local y. Defaults to None.
            char_y (float | None, optional): characteristic y-length of the element.
                    When not set, a characteristic value is computed from pos_field,
                    set to approximately the largest length curve of the position field
                    along constant local x. Defaults to None.

        Raises:
            DeformationGradient2DBadnessException: if the element is poorly shaped.

        Returns:
            tuple[tuple[float,float],bool]: ((x,y),success), where success is true
            if the error function is less than tol.
        """

        Np1 = self.num_nodes

        if char_x is None:
            # along each x-line, add the distances between nodes
            char_x = np.min(
                np.sum(  # take min across y-values, of x-line sums
                    np.linalg.norm(pos_field[1:, :, :] - pos_field[:-1, :, :], axis=-1),
                    axis=0,
                )
            )

        if char_y is None:
            char_y = np.min(
                np.sum(  # take min across x-values, of y-line sums
                    np.linalg.norm(pos_field[:, 1:, :] - pos_field[:, :-1, :], axis=-1),
                    axis=1,
                )
            )

        target = np.array((posx, posy))
        node_errs = np.sum((pos_field - target) ** 2, axis=-1)
        mindex = np.unravel_index(np.argmin(node_errs), (Np1, Np1))

        # local coords of guess
        local = np.array((self.knots[mindex[0]], self.knots[mindex[1]]))

        # position poly
        # sum(x^{i,j} L_{i,j}) -> [dim,k,l] (coefficient of cx^ky^l for dim)
        x_poly = np.einsum(
            "ijd,ik,jl->dkl", pos_field, self._lagrange_polys, self._lagrange_polys
        )

        # local to global
        def l2g(local):
            return np.einsum(
                "dkl,k,l->d",
                x_poly,
                local[0] ** np.arange(Np1),
                local[1] ** np.arange(Np1),
            )

        e = l2g(local) - target
        F = 0.5 * np.sum(e**2)

        # linsearch on gradient descent
        def linsearch(r, step):
            ARMIJO = 0.25
            err = l2g(local + r * step) - target
            F_new = 0.5 * np.sum(err**2)
            while F_new > F + (ARMIJO * step) * drF:
                step *= 0.5
                err[:] = l2g(local + r * step) - target
                F_new = 0.5 * np.sum(err**2)
            return F_new, step

        def clamp_and_maxstep(r):
            # out of bounds check; biggest possible step is to the boundary;
            # note: uses local[]. Additionally r[] is pass-by reference since it
            # can be modified. This is a feature, since that is the "clamp" part

            step = 1
            if ignore_out_of_bounds:
                return step
            for dim in range(2):  # foreach dim
                if (r[dim] < 0 and local[dim] == -1) or (
                    r[dim] > 0 and local[dim] == 1
                ):
                    # ensure descent direction does not point outside domain
                    # this allows constrained minimization
                    r[dim] = 0
                elif r[dim] != 0:
                    # prevent out-of-bounds step by setting maximum step;
                    if r[dim] > 0:
                        step = min(step, (1 - local[dim]) / r[dim])
                    else:
                        step = min(step, (-1 - local[dim]) / r[dim])
            return step

        iter_ = 0
        while tol < F and iter_ < max_iters:
            iter_ += 1

            # find descent direction by Newton

            # derivative of l2g
            dx_l2g = np.einsum(
                "dkl,k,l->d",
                x_poly[:, 1:, :],
                np.arange(1, Np1) * local[0] ** (np.arange(Np1 - 1)),
                local[1] ** np.arange(Np1),
            )
            dy_l2g = np.einsum(
                "dkl,k,l->d",
                x_poly[:, :, 1:],
                local[0] ** np.arange(Np1),
                np.arange(1, Np1) * local[1] ** (np.arange(Np1 - 1)),
            )

            # check badness
            defgrad_badness = np.abs(
                np.linalg.det([dx_l2g, dy_l2g]) / (char_x * char_y)  # type: ignore
            )
            if defgrad_badness < def_grad_badness_tol:
                raise DeformationGradient2DBadnessException(
                    defgrad_badness, local[0], local[1]
                )

            # grad of err function (ex^2 + ey^2)/2
            dxF = np.dot(e, dx_l2g)
            dyF = np.dot(e, dy_l2g)

            # solve hessian and descent dir
            hxx = np.sum(
                dx_l2g * dx_l2g
                + e
                * np.einsum(
                    "dkl,k,l->d",
                    x_poly[:, 2:, :],
                    np.arange(1, Np1 - 1)
                    * np.arange(2, Np1)
                    * local[0] ** (np.arange(Np1 - 2)),
                    local[1] ** (np.arange(Np1)),
                )
            )
            hxy = np.sum(
                dx_l2g * dy_l2g
                + e
                * np.einsum(
                    "dkl,k,l->d",
                    x_poly[:, 1:, 1:],
                    np.arange(1, Np1) * local[0] ** (np.arange(Np1 - 1)),
                    np.arange(1, Np1) * local[1] ** (np.arange(Np1 - 1)),
                )
            )
            hyy = np.sum(
                dy_l2g * dy_l2g
                + e
                * np.einsum(
                    "dkl,k,l->d",
                    x_poly[:, :, 2:],
                    local[0] ** (np.arange(Np1)),
                    np.arange(1, Np1 - 1)
                    * np.arange(2, Np1)
                    * local[1] ** (np.arange(Np1 - 2)),
                )
            )

            # target newton_rhaphson step
            r_nr = -np.linalg.solve([[hxx, hxy], [hxy, hyy]], [dxF, dyF])

            # target grad desc step
            r_gd = -np.array((dxF, dyF))
            r_gd /= np.linalg.norm(r_gd)  # scale gd step

            # take the better step between Newton-Raphson and grad desc
            s_nr = clamp_and_maxstep(r_nr)
            s_gd = clamp_and_maxstep(r_gd)

            # descent direction -- if dF . r == 0, we found minimum
            drF = dxF * r_gd[0] + dyF * r_gd[1]
            if drF > -dmin:
                break
            F_gd, s_gd = linsearch(r_gd, s_gd)

            # compare to NR
            F_nr = 0.5 * np.sum((l2g(local + r_nr * s_nr) - target) ** 2)
            if F_nr < F_gd:
                local += r_nr * s_nr
                e[:] = l2g(local) - target
                F = F_nr
            else:
                local += r_gd * s_gd
                e[:] = l2g(local) - target
                F = F_gd

            # nudge back in bounds in case of truncation error
            if not ignore_out_of_bounds:
                for dim in range(2):
                    local[dim] = min(1, max(-1, local[dim]))
        return (local, tol > F)

    def lagrange_eval1D(
        self,
        deriv_order: int,
        lag_index: ArrayLike | None = None,
        x: ArrayLike | None = None,
    ) -> NDArray:
        """
        Calculates the derivative
        $[(\\frac{\\partial d}{dx})^{deriv_order} L_{lag_index}(x)]_{x}$

        Note that "lagrange" refers to the lagrange interpolation polynomial,
        not lagrangian coordinates. This is a one-dimension helper function.

        deriv_order is taken as an integer.

        lag_index and x must be broadcastable to the same shape, following
        standard numpy broadcasting rules.

        Since the polynomial coefficient matrix is indexed by lag_index,
        that is, P[lag_index,:] is stored, it is advised that lag_index should
        not have more than one element per index. In other words, lag_index
        should be some subset, reshaping, and/or permutation of arange().

        Args:
            deriv_order (int): the order of the derivative to compute
            lag_index (ArrayLike | None, optional): an array of indices for sampling
                the Lagrange polynomials. If None, then
                `np.arange(degree+1)[:,...]` is used.
                Defaults to None.
            x (ArrayLike | None, optional): an array of points to sample the Lagrange
                polynomials. If None, then `self.knots` is used, and lag_index is
                treated as having an extra axis at the end. That is, the returned array
                has shape (*lag_index.shape, len(self.knots)). Defaults to None.

        Returns:
            NDArray: the result of the evaluation.
        """

        if x is None:
            if deriv_order not in self._lagrange_at_knots:
                self._lagrange_at_knots[deriv_order] = np.einsum(
                    "ik,jk->ij",
                    self.lagrange_poly1D(deriv_order),
                    self.knots[:, np.newaxis]
                    ** np.arange(self.num_nodes - deriv_order),
                )
            if lag_index is None:
                return self._lagrange_at_knots[deriv_order]
            # second index of arange should enforce the broadcastibility

            if not isinstance(lag_index, np.ndarray):
                lag_index = np.array(lag_index)
            return self._lagrange_at_knots[deriv_order][
                lag_index[..., np.newaxis], np.arange(self.num_nodes)
            ]
        if not isinstance(x, np.ndarray):
            x = np.array(x)
        if lag_index is None:
            lag_index = np.arange(self.num_nodes)[
                :, *(np.newaxis for _ in range(len(x.shape)))
            ]
        # out = sum_{k=deriv_order}^{degree}(
        #   k*(k-1)*...*(k-deriv_order+1) * L_poly[lag_index,k] * x^{k-deriv_order}
        #   )

        if not isinstance(lag_index, np.ndarray):
            lag_index = np.array(lag_index)
        return np.einsum(
            "...k,...k->...",
            self.lagrange_poly1D(deriv_order)[lag_index, :],
            np.expand_dims(x, -1) ** np.arange(self.num_nodes - deriv_order),
        )

    @typing.override
    def _compute_field_gradient(
        self,
        field: Field,
        pos_field: Field | None = None,
    ) -> Field:
        """
        Calculates the gradient of a field f, returning a new field.
        The result is an array of shape `(*basis_shape,...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (ArrayLike): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            pos_field (ArrayLike | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. This method supports element-stacking.
                Defaults to None.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.

        Returns:
            NDArray: an array representing the gradient of the field evaluated
                at each point.
        """

        x_deriv = fnp.einsum(
            ShapeComponent.BASIS,
            "ij,ia->aj",
            field,
            self.lagrange_eval1D(1),
        )
        y_deriv = fnp.einsum(
            ShapeComponent.BASIS,
            "ij,jb->ib",
            field,
            self.lagrange_eval1D(1),
        )
        grad = fnp.stack(
            Field.broadcast_fields_full(x_deriv, y_deriv),
            FieldAxisIndex(ShapeComponent.POINT, len(field.point_shape)),
        )

        if pos_field is not None:
            # (*pointshape,*stackshape,2{i},2{j}): dX^i/dx_j
            # must pad after stackshape
            def_grad = self.compute_field_gradient(pos_field)
            # grad is (*pointshape,*fieldshape,j): dF/dx_j
            # so we need the inverse
            return fnp.einsum(
                ShapeComponent.POINT,
                "ji,...j->...i",
                fnp.linalg.inv(def_grad),
                grad,
            )

        return grad

    @typing.override
    def _interpolate_field_gradient(
        self,
        field: Field,
        X: NDArray,
        Y: NDArray,
        pos_field: Field | None = None,
    ) -> Field:
        """
        Calculates the gradient of a field f at the reference coordinates (X,Y).
        The result is an array of shape `(...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        X and Y must have compatible shape.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (ArrayLike): an array of shape (*basis_shape,*fieldshape)
                representing the field to be interpolated.
            X (ArrayLike): x values (in reference coordinates).
            Y (ArrayLike): y values (in reference coordinates).
            pos_field (ArrayLike | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. Defaults to None.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.

        Returns:
            NDArray: an array representing the gradient of the field evaluated
                at each point.
        """

        x_deriv = fnp.einsum(
            ShapeComponent.BASIS,
            "ij,i...,j...->...",
            field,
            self.lagrange_eval1D(1, x=X),
            self.lagrange_eval1D(0, x=Y),
        )
        y_deriv = fnp.einsum(
            ShapeComponent.BASIS,
            "ij,i...,j...->...",
            field,
            self.lagrange_eval1D(0, x=X),
            self.lagrange_eval1D(1, x=Y),
        )
        grad = fnp.stack(
            Field.broadcast_fields_full(x_deriv, y_deriv),
            FieldAxisIndex(ShapeComponent.POINT, len(field.point_shape)),
        )

        if pos_field is not None:
            # (*pointshape,*stackshape*,2{i},2{j}): dX^i/dx_j
            def_grad = self.interpolate_field_gradient(pos_field, X, Y)
            # grad is (*pointshape,*fieldshape,j): dF/dx_j
            # so we need the inverse
            return fnp.einsum(
                ShapeComponent.POINT, "...ji,...j->...i", fnp.linalg.inv(def_grad), grad
            )

        return grad

    @typing.override
    def _integrate_field(
        self,
        pos_field: Field,
        field: Field,
        jacobian_scale: Field = Field((), (), 1),  # NOQA: B008
    ) -> Field:
        """
        Integrates `field` $f$ over the element. The result is the value of
        $\\int \\alpha f ~dV$ over the element, as an array of shape
        `(...,*point_shape)`.

        Args:
            pos_field (ArrayLike): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (ArrayLike): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            jacobian_scale (ArrayLike, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape `(...,*fieldshape)`.
        """
        Jw = fnp.einsum(
            ShapeComponent.BASIS,
            "i,j,ij,ij->ij",
            self.weights,
            self.weights,
            fnp.abs(fnp.linalg.det(self.compute_field_gradient(pos_field))),
            jacobian_scale,
        )
        return fnp.einsum(ShapeComponent.BASIS, "ij,ij->...", Jw, field)

    @typing.override
    def _integrate_basis_times_field(
        self,
        pos_field: Field,
        field: Field,
        indices: colltypes.Sequence[ArrayLike] | None = None,
        jacobian_scale: Field = Field((), (), 1),  # NOQA: B008
    ) -> Field:
        """
        Computes the integral $\\int \\alpha \\phi_i f ~ dV$
        for fields $f$ and $\\alpha$, on this element, given by `field` and
        `jacobian_scale`, respectively. When None, `jacobian_scale` is interpreted
        as the identity.

        Args:
            pos_field (ArrayLike): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (ArrayLike): an array of shape `(*basis_shape,...,*fieldshape)`
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (ArrayLike, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        Jw = fnp.einsum(
            ShapeComponent.BASIS,
            "i,j,ij,ij->ij",
            self.weights,
            self.weights,
            fnp.abs(fnp.linalg.det(self.compute_field_gradient(pos_field))),
            jacobian_scale,
        )
        if indices is None:
            return fnp.einsum(ShapeComponent.BASIS, "ij,ij->ij", Jw, field)
        return fnp.einsum(ShapeComponent.BASIS, "ij,ij->ij", Jw, field).basis[
            *indices, ...
        ]

    @typing.override
    def _mass_matrix(
        self,
        pos_field: Field,
        indices: colltypes.Sequence[np.ndarray] | None = None,
        jacobian_scale: Field = Field((), (), 1),  # NOQA: B008
    ) -> Field:
        """
        Recovers the mass matrix entries $\\int \\alpha \\phi_i \\phi_j ~ dV$
        for this element. `pos_field` has shape `(basis_shape,...,2)`, where the
        ellipses represent element stacking.

        The full mass matrix is of shape
        `(*basis_shape, *basis_shape,...)`, where the last indices match the element
        stack, but slices of the mass matrix can be
        obtained by passing into the `indices` argument. The implementation of
        `basis_mass_matrix` should assume that `indices` argument is smaller than the
        whole matrix, so if `M[*indices]` is larger than `M`, one should use
        `basis_mass_matrix(pos_field)[*indices,...]` instead of
        `basis_mass_matrix(pos_field,indices)`.

        Args:
            pos_field (ArrayLike): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the mass
                    matrix to access, or None if the whole matrix should be returned.
                    Defaults to None.
            jacobian_scale (ArrayLike, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: an array representing the mass matrix, or portions thereof.
        """
        # compute using GLL quadrature

        # int (phi1 * phi2) = sum_i(phi1(x_i) phi2(x_i) w_i * J(x_i))

        # weights [i,j] times jacobian

        Jw = fnp.einsum(
            ShapeComponent.BASIS,
            "i,j,ij,ij->ij",
            self.weights,
            self.weights,
            fnp.abs(fnp.linalg.det(self.compute_field_gradient(pos_field))),
            jacobian_scale,
        )
        basis_dims = len(self.basis_shape())
        if indices is None:
            double_basis = fnp.moveaxis(
                fnp.moveaxis(
                    self.basis_fields(),
                    (ShapeComponent.STACK, 1),
                    (ShapeComponent.BASIS, 0),
                ),
                (ShapeComponent.STACK, 0),
                (ShapeComponent.BASIS, 0),
            )
            return fnp.einsum(
                ShapeComponent.BASIS, "ijkl,...kl->ijkl", double_basis, Jw
            )
        indsI = np.array(indices[:basis_dims], dtype=int)
        indsJ = np.array(indices[basis_dims:], dtype=int)
        # TODO create fnp.where
        aux_pad = tuple(
            np.newaxis
            for _ in range(
                len(Jw.shape) - Jw._axis_field_to_numpy((ShapeComponent.BASIS, -1)) - 1
            )
        )
        Jw = Jw.basis[*indsJ]
        return Field(
            indsJ.shape,
            (),
            np.where(
                np.logical_and(
                    indsI[0, ...] == indsJ[0, ...], indsI[1, ...] == indsJ[1, ...]
                )[..., *aux_pad],
                Jw.coefficients,
                0,
            ),
            shape_order=Jw.shape_order,
        )

    @typing.override
    def _integrate_grad_basis_dot_field(
        self,
        pos_field: Field,
        field: Field,
        indices: colltypes.Sequence[ArrayLike] | None = None,
        jacobian_scale: Field = Field((), (), 1),  # NOQA: B008
    ) -> Field:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot f ~ dV$
        for a field $f$, on this element. The dot product takes the last axis of
        `field`, which must have size equal to the domain dimension.

        Args:
            pos_field (ArrayLike): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (ArrayLike): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            is_field_upper_index (bool): True if the last axis of field is treated as
                an upper (contravariant / vector) index. False if it is a lower
                (covariant / covector) index.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (ArrayLike, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """

        def_grad = self.compute_field_gradient(pos_field)  # ^g_l
        # int (grad phi1 . grad field)
        w = fnp.einsum(
            ShapeComponent.BASIS,
            "i,j,ij,ij->ij",
            self.weights,
            self.weights,
            fnp.abs(fnp.linalg.det(def_grad)),
            jacobian_scale,
        )
        # [k,i] L_k'(x_i)
        lag_div = self.lagrange_eval1D(1)
        # [i,j,...,dim] partial_dim field(xi,xj)

        def_grad_inv = fnp.linalg.inv(def_grad)  # ^l_g
        field_local_upper = fnp.einsum(
            ShapeComponent.POINT,
            "lg,...g->...l",
            def_grad_inv,
            field,
        )

        # integrand is
        # [ field(xi,xj) ] * [ partial_dim (L_m(xi)L_n(xj)) ]
        # where partial_dim (L_m(xi)L_n(xj)) = {dim==0: L_m'(xi)L_n(xj), dim==1: L_m(xi)L_n'(xj)}
        #                  = delta_{dim,0} delta_{nj} L_m'(xi) + delta_{dim,1} delta_{mi} L_n'(xj)

        # full integral is
        # [ field(xi,xj) ] * [ partial_dim (L_m(xi)L_n(xj)) ] * w_{ij}
        # = field(xi,xj) (delta_{dim,0} delta_{nj} L_m'(xi) + delta_{dim,1} delta_{mi} L_n'(xj)) w_{ij}
        # = (field(xi,xj)) delta_{dim,0} delta_{nj} L_m'(xi)w_{ij}
        #           + (field(xi,xj)) delta_{dim,1} delta_{mi} L_n'(xj) w_{ij}
        # = (field_0(xi,xn)) L_m'(xi)w_{in} + (field_1(xm,xj)) L_n'(xj) w_{mj}

        KF = fnp.einsum(
            ShapeComponent.BASIS,
            "in,mi,in->mn",
            field_local_upper.point[..., 0],
            lag_div,
            w,
        ) + fnp.einsum(
            ShapeComponent.BASIS,
            "mj,nj,mj->mn",
            field_local_upper.point[..., 1],
            lag_div,
            w,
        )
        if indices is None:
            return KF
        return KF.basis[*indices, ...]

    @typing.override
    def _integrate_grad_basis_dot_grad_field(
        self,
        pos_field: Field,
        field: Field,
        indices: colltypes.Sequence[ArrayLike] | None = None,
        jacobian_scale: Field = Field((), (), 1),  # NOQA: B008
    ) -> Field:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot \\nabla f ~ dV$
        for a field $f$, on this element. The dot product is over the gradients.

        Args:
            pos_field (ArrayLike): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (ArrayLike): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...]): the shape of `field` pointwise.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (ArrayLike, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        # compute using GLL quadrature

        def_grad = self.compute_field_gradient(pos_field)
        # int (grad phi1 . grad field)
        w = fnp.einsum(
            ShapeComponent.BASIS,
            "i,j,ij,ij->ij",
            self.weights,
            self.weights,
            fnp.abs(fnp.linalg.det(def_grad)),
            jacobian_scale,
        )
        # [k,i] L_k'(x_i)
        lag_div = self.lagrange_eval1D(1)
        # [i,j,...,dim] partial_dim field(xi,xj)

        grad_field_form = self.compute_field_gradient(field)
        def_grad_inv = fnp.linalg.inv(def_grad)
        grad_field = fnp.einsum(
            ShapeComponent.POINT,
            "ab,db,...d->...a",
            def_grad_inv,
            def_grad_inv,
            grad_field_form,
        )

        # integrand is
        # [ partial_dim field(xi,xj) ] * [ partial_dim (L_m(xi)L_n(xj)) ]
        # where partial_dim (L_m(xi)L_n(xj)) = {dim==0: L_m'(xi)L_n(xj), dim==1: L_m(xi)L_n'(xj)}
        #                  = delta_{dim,0} delta_{nj} L_m'(xi) + delta_{dim,1} delta_{mi} L_n'(xj)

        # full integral is
        # [ partial_dim field(xi,xj) ] * [ partial_dim (L_m(xi)L_n(xj)) ] * w_{ij}
        # = partial_dim field(xi,xj) (delta_{dim,0} delta_{nj} L_m'(xi) + delta_{dim,1} delta_{mi} L_n'(xj)) w_{ij}
        # = (partial_dim field(xi,xj)) delta_{dim,0} delta_{nj} L_m'(xi)w_{ij}
        #           + (partial_dim field(xi,xj)) delta_{dim,1} delta_{mi} L_n'(xj) w_{ij}
        # = (partial_0 field(xi,xn)) L_m'(xi)w_{in} + (partial_1 field(xm,xj)) L_n'(xj) w_{mj}

        KF = fnp.einsum(
            ShapeComponent.BASIS, "in,mi,in->mn", grad_field.point[..., 0], lag_div, w
        ) + fnp.einsum(
            ShapeComponent.BASIS, "mj,nj,mj->mn", grad_field.point[..., 1], lag_div, w
        )
        if indices is None:
            return KF
        return KF.basis[*indices, ...]  # type: ignore



import abc
import collections.abc as colltypes

import numpy as np
from numpy.typing import ArrayLike, NDArray

import fastfem.fields.numpy_similes as fnp
from fastfem.elements.element import ElementBase, IsoparametricElement
from fastfem.fields.field import Field as FieldType
from fastfem.fields.field import ShapeComponent
from fastfem.fields.field import _is_compatible as shapes_compatible


class Element2D(ElementBase, IsoparametricElement):
    """
    Handles the management of element operations. Element classes define operations
    to compute integrals on them. The elements themselves are handled in a data-driven
    way, where positions and fields are handled as arrays of coefficients of the shape
    functions defined by the element class. As isoparametric elements, positions
    and fields use the same set of shape functions. This class also assumes derivatives
    map into the same space, so integrate_field makes sense.
    """

    @abc.abstractmethod
    def reference_element_position_field(self) -> FieldType:
        """
        The position field of the reference (un-transformed) element. This is a vector
        field.

        Returns:
            Field: A field of shape `(basis_shape(),(),(2,))` corresponding to the
            position field.
        """
        raise NotImplementedError

    def interpolate_field(
        self,
        field: FieldType,
        X: ArrayLike,
        Y: ArrayLike,
    ) -> FieldType:
        """Evaluates field at (X,Y) in reference coordinates.
        The result is an array of values `field(X,Y)`.

        X and Y must have a compatible shape with `field.stack_shape`.

        Args:
            field (Field): the field to interpolate
                representing the field to be interpolated.
            X (ArrayLike): x values (in reference coordinates).
            Y (ArrayLike): y values (in reference coordinates).

        Returns:
            NDArray: The interpolated values `field(X,Y)`
        """
        self._verify_field_compatibilities(field=field)
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        if not shapes_compatible(np.shape(X), np.shape(Y)):
            message = f"shapes of X {np.shape(X)} and Y {np.shape(Y)} are incompatible."
            raise ValueError(message)
        field, _ = FieldType.broadcast_field_compatibility(
            field, FieldType(self.basis_shape(), (), np.array(0))
        )
        return self._interpolate_field(field, X, Y)

    @abc.abstractmethod
    def _interpolate_field(
        self,
        field: FieldType,
        X: NDArray,
        Y: NDArray,
    ) -> FieldType:
        """Evaluates field at (X,Y) in reference coordinates.
        The result is an array of values `field(X,Y)`.

        X and Y must have a compatible shape with `field.stack_shape`.

        Args:
            field (Field): the field to interpolate
                representing the field to be interpolated.
            X (NDArray): x values (in reference coordinates).
            Y (NDArray): y values (in reference coordinates).

        Returns:
            NDArray: The interpolated values `field(X,Y)`
        """
        raise NotImplementedError

    def compute_field_gradient(
        self,
        field: FieldType,
        pos_field: FieldType | None = None,
    ) -> FieldType:
        """
        Calculates the gradient of a field f, returning a new field.
        The result is an array of shape `(*basis_shape,...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            pos_field (Field | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. This method supports element-stacking.
                Defaults to None.

        Returns:
            Field: The field of the gradient.
        """
        if pos_field is None:
            self._verify_field_compatibilities(field=field)
            field, _ = FieldType.broadcast_field_compatibility(
                field, FieldType(self.basis_shape(), (), np.array(0))
            )
        else:
            self._verify_field_compatibilities(field=field, pos_field=pos_field)
            if pos_field.point_shape != (2,):
                message = (
                    "pos_field must have point_shape (2,). Found"
                    f" {pos_field.point_shape} instead."
                )
                raise ValueError(message)
            pos_field, field, _ = FieldType.broadcast_field_compatibility(
                pos_field, field, FieldType(self.basis_shape(), (), np.array(0))
            )

        return self._compute_field_gradient(field, pos_field)

    @abc.abstractmethod
    def _compute_field_gradient(
        self,
        field: FieldType,
        pos_field: FieldType | None,
    ) -> FieldType:
        """
        Calculates the gradient of a field f, returning a new field.
        The result is an array of shape `(*basis_shape,...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            pos_field (Field | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. This method supports element-stacking.
                Defaults to None.

        Returns:
            Field: The field of the gradient.
        """
        raise NotImplementedError

    def interpolate_field_gradient(
        self,
        field: FieldType,
        X: ArrayLike,
        Y: ArrayLike,
        pos_field: FieldType | None = None,
    ) -> FieldType:
        """
        Calculates the gradient of a field f at the reference coordinates (X,Y).
        The result is an array of shape `(...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        X and Y must have compatible shape.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (Field): an array of shape (*basis_shape,*fieldshape)
                representing the field to be interpolated.
            X (ArrayLike): x values (in reference coordinates).
            Y (ArrayLike): y values (in reference coordinates).
            pos_field (Field | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. Defaults to None.

        Returns:
            NDArray: an array representing the gradient of the field evaluated
                at each point.
        """

        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        if pos_field is None:
            self._verify_field_compatibilities(field=field)
            if not shapes_compatible(np.shape(X), np.shape(Y)):
                message = (
                    f" shapes of X {np.shape(X)} and Y {np.shape(Y)} incompatible."
                )
                raise ValueError(message)
            field, _ = FieldType.broadcast_field_compatibility(
                field, FieldType(self.basis_shape(), (), np.array(0))
            )
        else:
            self._verify_field_compatibilities(field=field, pos_field=pos_field)
            if not shapes_compatible(np.shape(X), np.shape(Y)):
                message = f"shapes of X {np.shape(X)} and Y {np.shape(Y)} incompatible."
                raise ValueError(message)
            if pos_field.point_shape != (2,):
                message = (
                    "pos_field must have point_shape (2,). Found"
                    f" {pos_field.point_shape} instead."
                )
                raise ValueError(message)
            pos_field, field, _ = FieldType.broadcast_field_compatibility(
                pos_field, field, FieldType(self.basis_shape(), (), np.array(0))
            )

        return self._interpolate_field_gradient(field, X, Y, pos_field)

    def _interpolate_field_gradient(
        self,
        field: FieldType,
        X: NDArray,
        Y: NDArray,
        pos_field: FieldType | None,
    ) -> FieldType:
        """
        Calculates the gradient of a field f at the reference coordinates (X,Y).
        The result is an array of shape `(...,*fieldshape,2)`,
        where field is of shape `(*basis_shape,...,*fieldshape)`.

        X and Y must have compatible shape.

        The last index is the coordinate of the derivative.

        This gradient can be computed in either reference space (with respect
        to the coordinates of the reference element), or in global space.
        If a global cartesian gradient should be calculated, then pos_field
        must be set to the coordinate matrix of the element. Otherwise,
        pos_field can be kept None.

        Args:
            field (Field): an array of shape (*basis_shape,*fieldshape)
                representing the field to be interpolated.
            X (NDArray): x values (in reference coordinates).
            Y (NDArray): y values (in reference coordinates).
            pos_field (Field | None, optional): If set, `pos_field` specifies
                the position fields of the element, and the gradient will be computed in
                Cartesian coordinates. Defaults to None.

        Returns:
            NDArray: an array representing the gradient of the field evaluated
                at each point.
        """
        message = (
            "Element._interpolate_field_gradient() called, which delegates to"
            " compute_field_gradient() and interpolate_field()"
        )
        self._dev_warn(message)
        return self.interpolate_field(
            self.compute_field_gradient(field, pos_field),
            X,
            Y,
        )

    def integrate_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Integrates `field` $f$ over the element. The result is the value of
        $\\int \\alpha f ~dV$ over the element, as an array of shape
        `(...,*point_shape)`.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape `(...,*fieldshape)`.
        """
        self._verify_field_compatibilities(
            pos_field=pos_field, field=field, jacobian_scale=jacobian_scale
        )
        if pos_field.point_shape != (2,):
            message = (
                "pos_field must have point_shape (2,). Found"
                f" {pos_field.point_shape} instead."
            )
            raise ValueError(message)
        pos_field, field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            pos_field,
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_field(pos_field, field, jacobian_scale)

    @abc.abstractmethod
    def _integrate_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Integrates `field` $f$ over the element. The result is the value of
        $\\int \\alpha f ~dV$ over the element, as an array of shape
        `(...,*point_shape)`.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape `(...,*fieldshape)`.
        """
        raise NotImplementedError

    def integrate_basis_times_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\phi_i f ~ dV$
        for fields $f$ and $\\alpha$, on this element, given by `field` and
        `jacobian_scale`, respectively. When None, `jacobian_scale` is interpreted
        as the identity.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape `(*basis_shape,...,*fieldshape)`
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the
                    basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(
            pos_field=pos_field, field=field, jacobian_scale=jacobian_scale
        )
        if pos_field.point_shape != (2,):
            message = (
                "pos_field must have point_shape (2,). Found"
                f" {pos_field.point_shape} instead."
            )
            raise ValueError(message)
        pos_field, field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            pos_field,
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_basis_times_field(
            pos_field, field, indices, jacobian_scale
        )

    @abc.abstractmethod
    def _integrate_basis_times_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\phi_i f ~ dV$
        for fields $f$ and $\\alpha$, on this element, given by `field` and
        `jacobian_scale`, respectively. When None, `jacobian_scale` is interpreted
        as the identity.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape `(*basis_shape,...,*fieldshape)`
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the
                    basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        raise NotImplementedError

    def mass_matrix(
        self,
        pos_field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Recovers the mass matrix entries $\\int \\alpha \\phi_i \\phi_j ~ dV$
        for this element. `pos_field` has shape `(basis_shape,...,2)`, where the
        ellipses represent element stacking.

        The full mass matrix is of shape
        `(*basis_shape, *basis_shape,...)`, where the last indices match the element
        stack, but slices of the mass matrix can be
        obtained by passing into the `indices` argument. The implementation of
        `basis_mass_matrix` should assume that `indices` argument is smaller than the
        whole matrix, so if `M[*indices]` is larger than `M`, one should use
        `basis_mass_matrix(pos_field)[*indices,...]` instead of
        `basis_mass_matrix(pos_field,indices)`.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the mass
                    matrix to access, or None if the whole matrix should be returned.
                    Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: an array representing the mass matrix, or portions thereof.
        """
        self._verify_field_compatibilities(
            pos_field=pos_field, jacobian_scale=jacobian_scale
        )
        if pos_field.point_shape != (2,):
            message = (
                "pos_field must have point_shape (2,). Found"
                f" {pos_field.point_shape} instead."
            )
            raise ValueError(message)
        pos_field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            pos_field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._mass_matrix(pos_field, indices, jacobian_scale)

    def _mass_matrix(
        self,
        pos_field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Recovers the mass matrix entries $\\int \\alpha \\phi_i \\phi_j ~ dV$
        for this element. `pos_field` has shape `(basis_shape,...,2)`, where the
        ellipses represent element stacking.

        The full mass matrix is of shape
        `(*basis_shape, *basis_shape,...)`, where the last indices match the element
        stack, but slices of the mass matrix can be
        obtained by passing into the `indices` argument. The implementation of
        `basis_mass_matrix` should assume that `indices` argument is smaller than the
        whole matrix, so if `M[*indices]` is larger than `M`, one should use
        `basis_mass_matrix(pos_field)[*indices,...]` instead of
        `basis_mass_matrix(pos_field,indices)`.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the mass
                    matrix to access, or None if the whole matrix should be returned.
                    Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: an array representing the mass matrix, or portions thereof.
        """
        self._dev_warn(
            "Element._mass_matrix() called, which delegates "
            "to integrate_basis_times_field()"
        )
        mat = fnp.moveaxis(
            self.integrate_basis_times_field(
                pos_field,
                fnp.moveaxis(
                    self.basis_fields(),
                    (ShapeComponent.STACK, 0),
                    (ShapeComponent.POINT, 0),
                ),
                None if indices is None else indices[: len(self.basis_shape())],
                jacobian_scale,
            ),
            (ShapeComponent.POINT, 0),
            (ShapeComponent.BASIS, 0),
        )
        return (
            mat if indices is None else mat.basis[*indices[len(self.basis_shape()) :]]
        )

    def integrate_grad_basis_dot_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot f ~ dV$
        for a field $f$, on this element. The dot product takes the last axis of
        `field`, which must have size equal to the domain dimension.
        The basis for the vector field is assumed to be in global coordinates,
        which is assumed to be over the kronecker dot product.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(
            pos_field=pos_field, field=field, jacobian_scale=jacobian_scale
        )
        if pos_field.point_shape != (2,):
            message = (
                "pos_field must have point_shape (2,). Found"
                f" {pos_field.point_shape} instead."
            )
            raise ValueError(message)
        pos_field, field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            pos_field,
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_grad_basis_dot_field(
            pos_field, field, indices, jacobian_scale
        )

    @abc.abstractmethod
    def _integrate_grad_basis_dot_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot f ~ dV$
        for a field $f$, on this element. The dot product takes the last axis of
        `field`, which must have size equal to the domain dimension.
        The basis for the vector field is assumed to be in global coordinates,
        which is assumed to be over the kronecker dot product.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        raise NotImplementedError

    def integrate_grad_basis_dot_grad_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot \\nabla f ~ dV$
        for a field $f$, on this element. The dot product is over the gradients.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(
            pos_field=pos_field, field=field, jacobian_scale=jacobian_scale
        )
        if pos_field.point_shape != (2,):
            message = (
                "pos_field must have point_shape (2,). Found"
                f" {pos_field.point_shape} instead."
            )
            raise ValueError(message)
        pos_field, field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            pos_field,
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_grad_basis_dot_grad_field(
            pos_field, field, indices, jacobian_scale
        )

    def _integrate_grad_basis_dot_grad_field(
        self,
        pos_field: FieldType,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot \\nabla f ~ dV$
        for a field $f$, on this element. The dot product is over the gradients.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._dev_warn(
            "Element._integrate_grad_basis_dot_grad_field() called, "
            "which delegates to integrate_grad_basis_dot_field()"
        )
        return self.integrate_grad_basis_dot_field(
            pos_field,
            self.compute_field_gradient(field, pos_field),
            indices,
            jacobian_scale,
        )

    # TODO boundary integrals,


class StaticElement2D(ElementBase, IsoparametricElement):
    """
    Handles the management of element operations. Element classes define operations
    to compute integrals on them. The elements themselves are handled in a data-driven
    way, where positions and fields are handled as arrays of coefficients of the shape
    functions defined by the element class. As isoparametric elements, positions
    and fields use the same set of shape functions, but it is not assumed that
    derivatives do as well.

    Static elements have a defined position field, which must be set at initialization.
    """

    position_field: FieldType

    def interpolate_field(
        self,
        field: FieldType,
        X: ArrayLike,
        Y: ArrayLike,
    ) -> FieldType:
        """Evaluates field at (X,Y) in reference coordinates.
        The result is an array of values `field(X,Y)`.

        X and Y must have a compatible shape with `field.stack_shape`.

        Args:
            field (Field): the field to interpolate
                representing the field to be interpolated.
            X (ArrayLike): x values (in reference coordinates).
            Y (ArrayLike): y values (in reference coordinates).

        Returns:
            FieldType: The interpolated values `field(X,Y)`
        """
        self._verify_field_compatibilities(field=field)
        if not isinstance(X, np.ndarray):
            X = np.array(X)
        if not isinstance(Y, np.ndarray):
            Y = np.array(Y)
        if not shapes_compatible(field.stack_shape, np.shape(X), np.shape(Y)):
            message = (
                f"field stack_shape {field.stack_shape} must be compatible with the"
                f" shapes of X {np.shape(X)} and Y {np.shape(Y)}."
            )
            raise ValueError(message)
        field, _ = FieldType.broadcast_field_compatibility(
            field, FieldType(self.basis_shape(), (), np.array(0))
        )
        return self._interpolate_field(field, X, Y)

    @abc.abstractmethod
    def _interpolate_field(
        self,
        field: FieldType,
        X: NDArray,
        Y: NDArray,
    ) -> FieldType:
        """Evaluates field at (X,Y) in reference coordinates.
        The result is an array of values `field(X,Y)`.

        X and Y must have a compatible shape with `field.stack_shape`.

        Args:
            field (Field): the field to interpolate
                representing the field to be interpolated.
            X (NDArray): x values (in reference coordinates).
            Y (NDArray): y values (in reference coordinates).

        Returns:
            FieldType: The interpolated values `field(X,Y)`
        """
        raise NotImplementedError

    def integrate_field(
        self,
        field: FieldType,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Integrates `field` $f$ over the element. The result is the value of
        $\\int \\alpha f ~dV$ over the element, as an array of shape
        `(...,*point_shape)`.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape `(...,*fieldshape)`.
        """
        self._verify_field_compatibilities(field=field, jacobian_scale=jacobian_scale)
        field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_field(field, jacobian_scale)

    @abc.abstractmethod
    def _integrate_field(
        self,
        field: FieldType,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Integrates `field` $f$ over the element. The result is the value of
        $\\int \\alpha f ~dV$ over the element, as an array of shape
        `(...,*point_shape)`.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape `(...,*fieldshape)`.
        """
        raise NotImplementedError

    def integrate_basis_times_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\phi_i f ~ dV$
        for fields $f$ and $\\alpha$, on this element, given by `field` and
        `jacobian_scale`, respectively. When None, `jacobian_scale` is interpreted
        as the identity.

        Args:
            field (Field): an array of shape `(*basis_shape,...,*fieldshape)`
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the
                    basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(field=field, jacobian_scale=jacobian_scale)
        field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_basis_times_field(field, indices, jacobian_scale)

    @abc.abstractmethod
    def _integrate_basis_times_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\phi_i f ~ dV$
        for fields $f$ and $\\alpha$, on this element, given by `field` and
        `jacobian_scale`, respectively. When None, `jacobian_scale` is interpreted
        as the identity.

        Args:
            field (Field): an array of shape `(*basis_shape,...,*fieldshape)`
                representing the field to be interpolated.
            fieldshape (tuple[int,...], optional): the shape of `field` pointwise.
                Defaults to tuple(), representing a scalar field.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the
                    basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        raise NotImplementedError

    def mass_matrix(
        self,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Recovers the mass matrix entries $\\int \\alpha \\phi_i \\phi_j ~ dV$
        for this element. `pos_field` has shape `(basis_shape,...,2)`, where the
        ellipses represent element stacking.

        The full mass matrix is of shape
        `(*basis_shape, *basis_shape,...)`, where the last indices match the element
        stack, but slices of the mass matrix can be
        obtained by passing into the `indices` argument. The implementation of
        `basis_mass_matrix` should assume that `indices` argument is smaller than the
        whole matrix, so if `M[*indices]` is larger than `M`, one should use
        `basis_mass_matrix(pos_field)[*indices,...]` instead of
        `basis_mass_matrix(pos_field,indices)`.

        Args:
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the mass
                    matrix to access, or None if the whole matrix should be returned.
                    Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: an array representing the mass matrix, or portions thereof.
        """
        self._verify_field_compatibilities(jacobian_scale=jacobian_scale)
        jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._mass_matrix(indices, jacobian_scale)

    def _mass_matrix(
        self,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Recovers the mass matrix entries $\\int \\alpha \\phi_i \\phi_j ~ dV$
        for this element. `pos_field` has shape `(basis_shape,...,2)`, where the
        ellipses represent element stacking.

        The full mass matrix is of shape
        `(*basis_shape, *basis_shape,...)`, where the last indices match the element
        stack, but slices of the mass matrix can be
        obtained by passing into the `indices` argument. The implementation of
        `basis_mass_matrix` should assume that `indices` argument is smaller than the
        whole matrix, so if `M[*indices]` is larger than `M`, one should use
        `basis_mass_matrix(pos_field)[*indices,...]` instead of
        `basis_mass_matrix(pos_field,indices)`.

        Args:
            pos_field (Field): an array representing the positions of the
                    element nodes. This is of the shape `(basis_shape,...,2)`
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the mass
                    matrix to access, or None if the whole matrix should be returned.
                    Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: an array representing the mass matrix, or portions thereof.
        """

        self._dev_warn(
            "Element._mass_matrix() called, which delegates "
            "to integrate_basis_times_field()"
        )
        mat = self.integrate_basis_times_field(
            self.basis_fields(),
            None if indices is None else indices[: len(self.basis_shape())],
            jacobian_scale,
        )
        return (
            mat if indices is None else mat.basis[*indices[len(self.basis_shape()) :]]
        )

    def integrate_grad_basis_dot_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot f ~ dV$
        for a field $f$, on this element. The dot product takes the last axis of
        `field`, which must have size equal to the domain dimension.
        The basis for the vector field is assumed to be in global coordinates,
        which is assumed to be over the kronecker dot product.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(field=field, jacobian_scale=jacobian_scale)
        field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_grad_basis_dot_field(field, indices, jacobian_scale)

    @abc.abstractmethod
    def _integrate_grad_basis_dot_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot f ~ dV$
        for a field $f$, on this element. The dot product takes the last axis of
        `field`, which must have size equal to the domain dimension.
        The basis for the vector field is assumed to be in global coordinates,
        which is assumed to be over the kronecker dot product.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        raise NotImplementedError

    def integrate_grad_basis_dot_grad_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None = None,
        jacobian_scale: FieldType = FieldType((), (), 1),  # NOQA: B008
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot \\nabla f ~ dV$
        for a field $f$, on this element. The dot product is over the gradients.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        self._verify_field_compatibilities(field=field, jacobian_scale=jacobian_scale)
        field, jacobian_scale, _ = FieldType.broadcast_field_compatibility(
            field,
            jacobian_scale,
            FieldType(self.basis_shape(), (), np.array(0)),
        )
        return self._integrate_grad_basis_dot_grad_field(field, indices, jacobian_scale)

    @abc.abstractmethod
    def _integrate_grad_basis_dot_grad_field(
        self,
        field: FieldType,
        indices: colltypes.Sequence | None,
        jacobian_scale: FieldType,
    ) -> FieldType:
        """
        Computes the integral $\\int \\alpha \\nabla \\phi_i \\cdot \\nabla f ~ dV$
        for a field $f$, on this element. The dot product is over the gradients.

        Args:
            field (Field): an array of shape (*basis_shape,...,*fieldshape)
                representing the field to be interpolated.
            indices (colltypes.Sequence[ArrayLike] | None, optional): Indices of the basis
                    functions to integrate against, or None if the integral should be
                    computed against every basis field. Defaults to None.
            jacobian_scale (Field, optional): an array of shape
                `(*basis_shape,...)` representing the scalar factor $\\alpha$.
                Defaults to 1.

        Returns:
            NDArray: The resultant integral, an array of shape
                `(indices.shape,...,*fieldshape)`, or `(*basis_shape,...,*fieldshape)`
                if `indices` is None.
        """
        raise NotImplementedError

    # TODO boundary integrals,



import time

import numpy as np
import pyvista as pv
from pyvista import CellType

import fastfem.mesh as m


class VisualMesh:
    def __init__(self, mesh: m.Mesh) -> None:
        """
        Initialize the plotter with the mesh.


        Args:
            mesh: The mesh object.
        """
        self.mesh = mesh

    def define_plotter(self) -> pv.UnstructuredGrid:
        """
        Given the Mesh object, redefines it for PyVista plotting.

        Returns:
            grid: PyVista grid object.
        """
        # Recovering element type
        if self.mesh["surface"].mesh[0].type == "triangle":
            cell_type = CellType.TRIANGLE
            side_number = 3
        else:
            cell_type = CellType.QUAD
            side_number = 4

        # Defining the points and strips
        points = np.array(list(self.mesh["surface"].mesh[0].nodes.values()))
        strips = np.array(list(self.mesh["surface"].mesh[0].elements.values()))

        # 0-indexing the strips
        strips_flat = strips.ravel() - 1
        cells = np.insert(
            strips_flat, np.arange(0, len(strips_flat), side_number), side_number
        )
        cell_arr = np.array(cells, dtype=np.int32)
        cell_types = np.full(len(strips), cell_type, dtype=np.uint8)

        # Create the unstructured grid
        return pv.UnstructuredGrid(cell_arr, cell_types, points)

    def plot_mesh(
        self,
        point_label: bool = False,
        mesh_color: str = "white",
        edge_color: str = "black",
        edge_thickness: int = 1,
    ) -> None:
        """
        Plots the mesh


        Args:
            mesh_color: Color of the mesh.
            edge_color: Color of the edges.
            edge_thickness: Thickness of the edges.
            point_label: Boolean value to determine whether the points are labeled.
        """
        # Define the grid and plotter objects
        grid = self.define_plotter()
        plotter = pv.Plotter()
        plotter.add_mesh(
            grid,
            show_edges=True,
            color=mesh_color,
            edge_color=edge_color,
            line_width=edge_thickness,
        )

        # Add point labels, if specified
        points = grid.points

        if point_label:
            mask = points[:, 2] == 0  # Labeling points on xy plane
            formatted_labels = [f"({p[0]:.2f}, {p[1]:.2f})" for p in points[mask]]
            plotter.add_point_labels(
                points[mask], formatted_labels, point_size=20, font_size=10
            )
        plotter.camera_position = "xy"
        plotter.show()

    def plot_data(self, data: np.ndarray, cmap: str = "viridis") -> None:
        """
        Plots the mesh with temperature data, for a single time step.

        Args:
            data: The temperature data for each node, contained in a 2D array.
            cmap: Colormap for the data.
        """
        # Define the grid and plotter objects
        grid = self.define_plotter()
        plotter = pv.Plotter()

        # Assign temperature data to the grid
        grid.point_data["Data"] = data.flatten()
        plotter.add_mesh(
            mesh=grid,
            scalars="Data",
            cmap=cmap,
            scalar_bar_args={
                "vertical": False,
                "title": "Data",
                "title_font_size": 25,
                "label_font_size": 10,
            },
        )
        plotter.camera_position = "xy"
        plotter.show()

    def animate_data(
        self,
        fps: float,
        total_time: float,
        data: np.ndarray,
        force: bool = False,
        cmap: str = "viridis",
    ) -> None:
        """
        Plots the mesh with temperature data, for successive time steps.

        Args:
            fps: Frames per second for the animation.
            total_time: Total time for the animation.
            data: The temperature data for each node, contained in a 3D array.
            cmap: Colormap for the data.
        """
        if fps > 25 and not force:
            message = "The maximum value for fps is 25. Please decrease your fps value."
            raise ValueError(message)

        # Define the grid and plotter objects
        grid = self.define_plotter()
        plotter = pv.Plotter()

        # Defining the number of time steps
        frames = int(total_time * fps)

        # Assign temperature data to the grid
        grid.point_data["Data"] = data[0].flatten()
        plotter.add_mesh(
            mesh=grid,
            color="red",
            render_points_as_spheres=True,
            point_size=10,
            scalars="Data",
            cmap=cmap,
            scalar_bar_args={
                "title_font_size": 20,
                "label_font_size": 16,
                "n_labels": 3,
                "italic": True,
                "fmt": "%.2f",
                "font_family": "arial",
            },
        )

        # Setting plotter settings
        plotter.camera_position = "xy"
        plotter.show(auto_close=False, interactive_update=True)
        plotter.render()
        text_actor = None

        # Animate
        for i in range(1, frames):
            # Starting a timer, accounting for the time taken to render the frame
            start_time = time.time()

            # Update the data
            grid.point_data["Data"] = data[i].flatten()
            plotter.render()

            # Update time text
            if text_actor is not None:
                plotter.remove_actor(text_actor)  # type:ignore
            text_actor = plotter.add_text(
                f"t = {i / fps:.2f}", position="upper_left", font_size=26, color="black"
            )

            # Ending the timer, waiting accordingly
            end_time = time.time()
            remaining_time = 1 / fps - (end_time - start_time)
            if remaining_time > 0:
                time.sleep(remaining_time)

    def make_movie(
        self,
        filename: str,
        fps: float,
        total_time: float,
        data: np.ndarray,
        cmap: str = "viridis",
    ) -> None:
        """
        Creates a movie with the given mesh and temperature data.

        Args:
            filename: The name of the file to be created.
            fps: Frames per second for the movie.
            total_time: Total time for the movie.
            data: The temperature data for each node, contained in a 2D array.
            cmap: Colormap for the data.
        """
        # Define the grid and plotter objects
        grid = self.define_plotter()
        plotter = pv.Plotter(off_screen=True)

        # Defining the number of time steps
        frames = int(total_time * fps)

        # Assign initial temperature data
        grid.point_data["Data"] = data[0].flatten("F")
        plotter.add_mesh(grid, scalars="Data", cmap=cmap)
        plotter.camera_position = "xy"

        # Creating file
        plotter.open_movie(filename)
        # plotter.show()
        plotter.show(auto_close=False)
        plotter.write_frame()
        text = None

        # Animate and record frames
        for i in range(1, frames):
            # Update the data
            grid.point_data["Data"] = data[i].flatten("F")
            plotter.render()

            # Update time text
            if text is not None:
                plotter.remove_actor(text)  # type:ignore
            text = plotter.add_text(
                f"t = {i / fps:.2f}", position="upper_left", font_size=20
            )
            plotter.write_frame()

        plotter.close()

    def make_gif(
        self,
        filename: str,
        fps: float,
        total_time: float,
        data: np.ndarray,
        cmap: str = "viridis",
    ) -> None:
        """
        Creates a GIF with the given mesh and temperature data.

        Args:
            filename: The name of the file to be created.
            fps: Frames per second for the movie.
            total_time: Total time for the movie.
            data: The temperature data for each node, contained in a 2D array.
            cmap: Colormap for the data.
        """
        # Define the grid and plotter objects
        grid = self.define_plotter()
        plotter = pv.Plotter(off_screen=True)

        # Defining the number of time steps
        frames = int(total_time * fps)

        # Assign initial temperature data
        grid.point_data["Data"] = data[0].flatten("F")
        plotter.add_mesh(grid, scalars="Data", cmap=cmap)
        plotter.camera_position = "xy"

        # Creating file
        plotter.open_gif(filename)
        # plotter.show(auto_close=False)
        # plotter.write_frame()
        text = None

        for i in range(1, frames):
            # Update the data
            grid.point_data["Data"] = data[i].flatten("F")
            plotter.render()

            # Update time text
            if text is not None:
                plotter.remove_actor(text)  # type:ignore
            text = plotter.add_text(
                f"t = {i / fps:.2f}", position="upper_left", font_size=20
            )
            plotter.write_frame()

        plotter.close()



import numpy as np
import pytest

import fastfem.fields.field as field


def shapes_generator(
    dimcountlims: tuple[int, int] = (0, 4),
    dimsizelims: tuple[int, int] = (0, 2),
    totalsizelims: tuple[int, int | None] = (0, 20),
):
    if (
        (dimcountlims[1] < dimcountlims[0] or dimcountlims[1] < 0)
        or (dimsizelims[1] < dimsizelims[0] or dimsizelims[1] < 0)
        or (
            totalsizelims[1] is not None and totalsizelims[1] < max(0, totalsizelims[0])
        )
    ):
        return
    if (dimcountlims[1] == 0) and (
        totalsizelims[0] <= 0 and (totalsizelims[1] is None or totalsizelims[1] >= 0)
    ):
        yield tuple()
        return

    yield from shapes_generator(
        (dimcountlims[0], dimcountlims[1] - 1), dimsizelims, totalsizelims
    )
    for shape in shapes_generator(
        (dimcountlims[1] - 1, dimcountlims[1] - 1), dimsizelims, totalsizelims
    ):
        for k in range(dimsizelims[0], dimsizelims[1] + 1):
            newshape = shape + (k,)
            size = np.prod(newshape, dtype=int)
            if totalsizelims[0] <= size and (
                totalsizelims[1] is None or size <= totalsizelims[1]
            ):
                yield newshape


def shapes():
    yield from shapes_generator(dimcountlims=(0, 3))
    yield from ((1,) * i + (3, 2) for i in range(3, 5))
    yield from ((1,) * i + (3, 2, 2) for i in range(3, 5))


def shapes_small():
    # yield from shapes_generator(dimcountlims=(0,2),dimsizelims=(1,2))
    # yield from ((1,)*i + (3,2) for i in range(3,5))
    # yield from ((1,)*i + (3,2,2) for i in range(3,5))
    # yield from shapes_generator(dimcountlims=(1,2),dimsizelims=(0,1))
    yield tuple()
    yield from ((i,) for i in range(3))
    yield from ((1,) * i + (3, 5) for i in range(2))
    yield from ((1,) * i + (5, 3) for i in range(2))
    yield (2, 1, 2)
    yield (0, 1)


def test_field_construction():
    # attr should fail for other things
    with pytest.raises(AttributeError):
        # this should fail
        field.Field(tuple(), tuple(), 1)._shape  # pyright: ignore
    for basis_shape in shapes_small():
        for point_shape in shapes_small():
            f = field.Field(
                basis_shape, point_shape, np.empty(point_shape)
            )  # this will work
            assert f.shape == (tuple(), basis_shape, point_shape)
            for stack_shape in shapes_small():
                coefs = np.empty(stack_shape + basis_shape + point_shape)
                f = field.Field(
                    basis_shape,
                    point_shape,
                    coefs,
                )  # this will work
                if f.shape != (stack_shape, basis_shape, point_shape):
                    f = field.Field(
                        basis_shape,
                        point_shape,
                        coefs,
                    )  # this will work

                assert f.shape == (stack_shape, basis_shape, point_shape)

                # these wont
                if np.prod(basis_shape, dtype=int) > 1:
                    basis_shape_ = tuple(ax + 1 for ax in basis_shape)
                    with pytest.raises(field.FieldConstructionError):
                        field.Field(basis_shape_, point_shape, coefs)
                if np.prod(point_shape, dtype=int) > 1:
                    point_shape_ = tuple(ax + 1 for ax in point_shape)
                    with pytest.raises(field.FieldConstructionError):
                        field.Field(basis_shape, point_shape_, coefs)


def test_shape_broadcastability_and_compatibility(comparison_three_shapes):
    for a, b, c, np_compatible in comparison_three_shapes:
        broadcastable = field._is_broadcastable(a, b, c)
        compatible = field._is_compatible(a, b, c)
        if np_compatible:
            target = np.broadcast_shapes(a, b, c)
            assert compatible, (
                f"Numpy broadcasts {a}, {b}, {c} -> {target}, but"
                f" _is_compatible(...) == {compatible}, which should be True."
            )
            assert (target == a) == broadcastable, (
                f"Numpy broadcasts {a}, {b}, {c} -> {target}, but"
                f" _is_broadcastable(...) == {compatible}, which should be"
                f" {target == a}."
            )
        else:
            assert not (broadcastable or compatible), (
                f"Numpy failed to broadcast {a}, {b}, {c}, but this disagrees"
                " with field functions: _is_broadcastable(...) =="
                f" {broadcastable}, _is_compatible(...) == {compatible}."
            )


def test_field_broadcast_compatibility_on_triples(comparison_three_shapes):
    for a, b, c, _ in comparison_three_shapes:
        for a_, b_, c_, _ in comparison_three_shapes:
            nonconst = [s for s in (a, b, c) if np.prod(s, dtype=int) != 1]
            compatibility = field._is_compatible(a_, b_, c_) and (
                len(nonconst) == 0 or all(nonconst[0] == s for s in nonconst)
            )
            fa = field.Field(a, tuple(), np.random.rand(*(a_ + a)))
            fb = field.Field(b, tuple(), np.random.rand(*(b_ + b)))
            fc = field.Field(c, tuple(), np.random.rand(*(c_ + c)))
            assert field.Field.are_compatible(fa, fb, fc) == compatibility, (
                f"{fa.shape}, {fb.shape} and"
                f" {fc.shape} broadcastibility should be"
                f" {compatibility}."
            )
            if compatibility:
                fa2, fb2, fc2 = field.Field.broadcast_field_compatibility(fa, fb, fc)
                assert fa == fa2
                assert fb == fb2
                assert fc == fc2
            else:
                with pytest.raises(field.FieldShapeError):
                    field.Field.broadcast_field_compatibility(fa, fb, fc)


def test_field_reshaping_errors():
    for a in shapes_small():
        for a_ in shapes_small():
            for a__ in shapes_small():
                fa = field.Field(a, a__, np.random.rand(*(a_ + a + a__)))
                if np.prod(a, dtype=int) > 1:
                    adiff = tuple(ax + 1 for ax in a)
                    with pytest.raises(field.FieldShapeError):
                        fa.broadcast_to_shape(a_, adiff, a__)
                if np.prod(a_, dtype=int) > 1:
                    a_diff = tuple(ax + 1 for ax in a_)
                    with pytest.raises(field.FieldShapeError):
                        fa.broadcast_to_shape(a_diff, a, a__)
                if np.prod(a__, dtype=int) > 1:
                    a__diff = tuple(ax + 1 for ax in a__)
                    with pytest.raises(field.FieldShapeError):
                        fa.broadcast_to_shape(a_, a, a__diff)


def test_field_broadcast_full_on_doubles(comparison_two_shapes):
    for val1, val2, _ in comparison_two_shapes:
        for a, b, a_, b_, a__, b__ in [
            (val1, val2, val1, val1, val1, val1),
            (val1, val1, val1, val2, val1, val1),
            (val1, val1, val1, val1, val1, val2),
        ]:
            broadcastibility = (
                field._is_compatible(a_, b_)
                and field._is_compatible(a__, b__)
                and (np.prod(b, dtype=int) == 1 or np.prod(a, dtype=int) == 1 or a == b)
            )
            fa = field.Field(a, a__, np.random.rand(*(a_ + a + a__)))
            fb = field.Field(b, b__, np.random.rand(*(b_ + b + b__)))
            assert field.Field.are_broadcastable(fa, fb) == broadcastibility, (
                f"{fa.shape} and {fb.shape} broadcastibility should be"
                f" {broadcastibility}."
            )
            if broadcastibility:
                fa2, fb2 = field.Field.broadcast_fields_full(fa, fb)
                assert fa == fa2
                assert fb == fb2
                fa3 = fa.broadcast_to_shape(
                    fa2.stack_shape, fa2.basis_shape, fa2.point_shape
                )
                assert fa3 == fa2
            else:
                assert fa != fb
                with pytest.raises(field.FieldShapeError):
                    field.Field.broadcast_fields_full(fa, fb)


def test_field_accessors():
    def random_accessors(n, base_shape):
        # yields n accessors for base_shape
        def rand_symbol(dimsize):
            if np.random.rand() < 0.5 and dimsize > 0:
                return np.random.randint(dimsize)
            return slice(None)

        shapesize = len(base_shape)
        if shapesize > 0:
            for _ in range(n):
                k = np.random.randint(shapesize)
                yield tuple(rand_symbol(base_shape[i]) for i in range(k))

    def prod_chain(A, B, C):
        return (A[..., *(None for _ in B.shape)] * B)[..., *(None for _ in C.shape)] * C

    for a in shapes_small():
        for b in shapes_small():
            for c in shapes_small():
                coefs_A = np.array(np.random.rand(*a))
                coefs_B = np.array(np.random.rand(*b))
                coefs_C = np.array(np.random.rand(*c))
                f = field.Field(b, c, prod_chain(coefs_A, coefs_B, coefs_C))
                if len(f.coefficients.shape) == 0:
                    continue
                for amod in random_accessors(5, a):
                    np.testing.assert_allclose(
                        f.stack[*amod].coefficients,
                        prod_chain(coefs_A[amod], coefs_B, coefs_C),
                    )
                for bmod in random_accessors(5, b):
                    np.testing.assert_allclose(
                        f.basis[*bmod].coefficients,
                        prod_chain(coefs_A, coefs_B[bmod], coefs_C),
                    )
                for cmod in random_accessors(5, c):
                    np.testing.assert_allclose(
                        f.point[*cmod].coefficients,
                        prod_chain(coefs_A, coefs_B, coefs_C[cmod]),
                    )


def test_axis_recovery(field_of_different_shape_orders_testsep):
    def validate_shapes(shapes):
        arr = np.empty(shapes[0] + shapes[1] + shapes[2])
        for f in field_of_different_shape_orders_testsep(shapes, arr):
            assert f._component_offset(f.shape_order_inverse[0]) == 0
            assert f._component_offset(f.shape_order_inverse[1]) == len(shapes[0])
            assert f._component_offset(f.shape_order_inverse[2]) == len(
                shapes[0]
            ) + len(shapes[1])

    validate_shapes([(2, 3, 2), (2, 2), (3, 3)])
    validate_shapes([(2, 3), (4, 1), (1, 2)])


def test_reshape_axis_down(field_of_different_shape_orders_testsep):
    shapes = [(2, 3, 2), (2, 3), (2, 3)]
    reshapes = [(12, 2, 3, 2, 3), (2, 3, 2, 6, 2, 3), (2, 3, 2, 2, 3, 6)]
    arr = np.random.rand(*shapes[0], *shapes[1], *shapes[2])
    for f in field_of_different_shape_orders_testsep(shapes, arr):
        ind = f.shape_order[field.ShapeComponent.STACK]
        np.testing.assert_allclose(
            f.stack.reshape(-1).coefficients, np.reshape(arr, reshapes[ind])
        )

        ind = f.shape_order[field.ShapeComponent.BASIS]
        np.testing.assert_allclose(
            f.basis.reshape(-1).coefficients, np.reshape(arr, reshapes[ind])
        )

        ind = f.shape_order[field.ShapeComponent.POINT]
        np.testing.assert_allclose(
            f.point.reshape(-1).coefficients, np.reshape(arr, reshapes[ind])
        )



import numpy as np

import fastfem.fields.numpy_similes as fnp
from fastfem.fields.field import Field, ShapeComponent


def test_moveaxis(permutation):
    def sum_shapes(shapes):
        return shapes[0] + shapes[1] + shapes[2]

    shapes = [(0, 1, 2), (3, 4, 5), (6, 7, 8)]
    perm_shapes = [()] * 3
    for i in range(len(perm_shapes)):
        perm_shapes[permutation[i]] = shapes[i]
    arr = np.broadcast_to(0, sum_shapes(perm_shapes))
    f = Field(
        shapes[ShapeComponent.BASIS],
        shapes[ShapeComponent.POINT],
        arr,
        shape_order=permutation,
    )

    def assert_cascade_down(p1, p2, s1, s2):
        shapes_swp = np.array(sum_shapes(shapes), dtype=int)
        rem = shapes_swp[s1]
        shapes_swp[s1:s2] = shapes_swp[s1 + 1 : s2 + 1]
        shapes_swp[s2] = rem
        moved = fnp.moveaxis(f, p1, p2).shape
        assert sum_shapes(moved) == tuple(int(k) for k in shapes_swp)
        shape_sizes = list(len(k) for k in shapes)
        shape_sizes[p1[0]] -= 1
        shape_sizes[p2[0]] += 1
        assert tuple(shape_sizes) == tuple(len(k) for k in moved)

    def assert_cascade_up(p1, p2, s1, s2):
        shapes_swp = np.array(sum_shapes(shapes), dtype=int)
        rem = shapes_swp[s2]
        shapes_swp[s1 + 1 : s2 + 1] = shapes_swp[s1:s2]
        shapes_swp[s1] = rem
        moved = fnp.moveaxis(f, p1, p2).shape
        assert sum_shapes(moved) == tuple(int(k) for k in shapes_swp)
        shape_sizes = list(len(k) for k in shapes)
        shape_sizes[p1[0]] -= 1
        shape_sizes[p2[0]] += 1
        assert tuple(shape_sizes) == tuple(len(k) for k in moved)

    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.STACK, 1), 0, 1)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.STACK, 2), 0, 2)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.STACK, -1), 0, 2)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 0), 0, 2)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 1), 0, 3)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 2), 0, 4)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 3), 0, 5)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.BASIS, -1), 0, 5)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.POINT, 0), 0, 5)
    assert_cascade_down((ShapeComponent.STACK, 0), (ShapeComponent.POINT, 1), 0, 6)
    assert_cascade_down((ShapeComponent.STACK, 1), (ShapeComponent.STACK, 0), 0, 1)
    assert_cascade_up((ShapeComponent.POINT, -1), (ShapeComponent.STACK, 0), 0, 8)



import numpy as np

import fastfem.mesh as m
import fastfem.plotter as p

mesh = m.create_a_rectangle_mesh(
    horizontal_length=2,
    vertical_length=1,
    nodes_in_horizontal_direction=20,
    nodes_in_vertical_direction=20,
    element_type="triangle",
    file_name=None,
)

# Data generation
TOTAL_TIME = 3
FPS = 25
TIME_STEPS = int(TOTAL_TIME * FPS)
DATA = np.random.uniform(
    0,
    50,
    size=(TIME_STEPS, 20, 20),
)

# Visualize
visualizer = p.VisualMesh(mesh)
visualizer.plot_mesh()
visualizer.animate_data(FPS, TOTAL_TIME, DATA)
visualizer.make_movie("filename_movie.mp4", FPS, TOTAL_TIME, DATA)
visualizer.make_gif("filename_gif.gif", FPS, TOTAL_TIME, DATA)


import fastfem.mesh as m

square_mesh = (
    m.create_a_square_mesh(
        side_length=1.0,
        nodes_in_horizontal_direction=10,
        nodes_in_vertical_direction=10,
        element_type="quadrangle",
    ),
)

create_a_rectangle_mesh = (
    m.create_a_rectangle_mesh(
        horizontal_length=1.0,
        vertical_length=0.5,
        nodes_in_horizontal_direction=10,
        nodes_in_vertical_direction=10,
        element_type="triangle",
    ),
)



from fastfem.fields.field import Field, FieldShapeError, ShapeComponent
from fastfem.fields.field import np_or_jnp as np


def det(
    field: Field, component_selector: ShapeComponent = ShapeComponent.POINT
) -> Field:
    """This is analogous to [`numpy.linalg.det`](https://numpy.org/doc/stable/reference/generated/numpy.linalg.det.html)

    Computes the determinant of a field along the given component. By default, the determinant is point-wise.
    Raises a FieldShapeError if that component is not a square matrix.

    Args:
        field (Field): The field to compute the determinant of
        component_selector (ShapeComponent, optional): The component to take the determinant over. Defaults to ShapeComponent.POINT.
    Raises:
        FieldShapeError: if the selected component is not square.

    Returns:
        Field: A field representing the determinant, with the selected component contracted.
    """
    selshape = field.get_shape(component_selector)
    if len(selshape) != 2 and selshape[0] != selshape[1]:
        message = (
            "Cannot take the determinant along component"
            f" {component_selector} ({selshape}). Must be square!"
        )
        raise FieldShapeError(message)
    shapes = [()] * 3
    shapes[ShapeComponent.BASIS] = field.basis_shape  # type: ignore
    shapes[ShapeComponent.POINT] = field.point_shape  # type: ignore
    shapes[ShapeComponent.STACK] = field.stack_shape  # type: ignore
    shapes[component_selector] = ()
    _np = np(field)
    return Field(
        shapes[ShapeComponent.BASIS],
        shapes[ShapeComponent.POINT],
        _np.linalg.det(
            _np.moveaxis(
                field.coefficients,
                (
                    field._axis_field_to_numpy((component_selector, 0)),
                    field._axis_field_to_numpy((component_selector, 1)),
                ),
                (-2, -1),
            )
        ),
        shape_order=field.shape_order,
        use_jax=field.use_jax,
    )


def inv(
    field: Field, component_selector: ShapeComponent = ShapeComponent.POINT
) -> Field:
    """This is analogous to [`numpy.linalg.inv`](https://numpy.org/doc/stable/reference/generated/numpy.linalg.inv.html)

    Computes the matrix inverse of a field along the given component. By default, the inverse is point-wise.
    Raises a FieldShapeError if that component is not a square matrix.

    Args:
        field (Field): The field to compute the matrix inverse of
        component_selector (ShapeComponent, optional): The component to take the inverse over. Defaults to ShapeComponent.POINT.
    Raises:
        FieldShapeError: if the selected component is not square.

    Returns:
        Field: A field representing the inverse, with the same shape.
    """
    selshape = field.get_shape(component_selector)
    if len(selshape) != 2 and selshape[0] != selshape[1]:
        message = (
            "Cannot take the determinant along component"
            f" {component_selector} ({selshape}). Must be square!"
        )
        raise FieldShapeError(message)
    _np = np(field)
    return Field(
        field.basis_shape,
        field.point_shape,
        _np.moveaxis(
            _np.linalg.inv(
                _np.moveaxis(
                    field.coefficients,
                    (
                        field._axis_field_to_numpy((component_selector, 0)),
                        field._axis_field_to_numpy((component_selector, 1)),
                    ),
                    (-2, -1),
                )
            ),
            (-2, -1),
            (
                field._axis_field_to_numpy((component_selector, 0)),
                field._axis_field_to_numpy((component_selector, 1)),
            ),
        ),
        shape_order=field.shape_order,
        use_jax=field.use_jax,
    )


__all__ = ["det", "inv"]



import numpy as np
import pytest

import fastfem.mesh as ffm
from fastfem.elements.linear_simplex2d import LinearSimplex2D
from fastfem.elements.linear_simplex_mesh2d import LinearSimplexMesh2D
from fastfem.fields.field import Field

meshes = {
    "rect1": ffm.create_a_rectangle_mesh(
        horizontal_length=1,
        vertical_length=1,
        nodes_in_horizontal_direction=10,
        nodes_in_vertical_direction=10,
        element_type="triangle",
        file_name=None,
    ),
    "rect2": ffm.create_a_rectangle_mesh(
        horizontal_length=1,
        vertical_length=2,
        nodes_in_horizontal_direction=None,
        nodes_in_vertical_direction=None,
        element_type="triangle",
        file_name=None,
    ),
}

LS = LinearSimplex2D()


@pytest.fixture(scope="session", params=meshes.keys())
def mesh(request):
    mesh = meshes[request.param]
    return mesh


@pytest.fixture(scope="session")
def element(mesh):
    return LinearSimplexMesh2D(mesh)


def test_atomstack(element):
    stackfield = element.to_atomstack(
        element.Field(np.arange(element.num_nodes), False)
    )
    assert stackfield.shape == ((element.num_elements,), (3,), tuple())
    # maybe want to validate that nodes are assembled correctly


def test_integrate_field(element):
    total = np.zeros((element.num_nodes,))
    for tri in element.element_node_indices:
        total[tri] += LS.integrate_field(
            Field((3,), (2,), element.node_coords[tri, :]),
            Field((3,), tuple(), np.eye(3)),
        ).coefficients
    np.testing.assert_allclose(
        element.integrate_field(element.basis_fields()).coefficients, total
    )


def test_integrate_basis_times_field(element):
    total = np.zeros((element.num_nodes, element.num_nodes))
    for tri in element.element_node_indices:
        total[tri[:, np.newaxis], tri] += LS.integrate_basis_times_field(
            Field((3,), (2,), element.node_coords[tri, :]),
            Field((3,), tuple(), np.eye(3)),
        ).coefficients
    np.testing.assert_allclose(
        element.integrate_basis_times_field(element.basis_fields()).coefficients, total
    )


def test_integrate_grad_basis_dot_field(element):
    total = np.zeros((element.num_nodes, element.num_nodes, 2))
    testfield = Field(
        (element.num_nodes,),
        (2, 2),
        element.basis_fields().coefficients[:, :, None, None] * np.eye(2),
    )
    for tri in element.element_node_indices:
        total[tri[:, np.newaxis], tri, :] += LS.integrate_grad_basis_dot_field(
            Field((3,), (2,), element.node_coords[tri, :]),
            Field((3,), (2, 2), np.eye(3)[:, :, None, None] * np.eye(2)),
        ).coefficients
    np.testing.assert_allclose(
        element.integrate_grad_basis_dot_field(testfield).coefficients,
        total,
    )


def test_integrate_grad_basis_dot_grad_field(element):
    total = np.zeros((element.num_nodes, element.num_nodes))
    for tri in element.element_node_indices:
        total[tri[:, np.newaxis], tri] += LS.integrate_grad_basis_dot_grad_field(
            Field((3,), (2,), element.node_coords[tri, :]),
            Field((3,), tuple(), np.eye(3)),
        ).coefficients
    np.testing.assert_allclose(
        element.integrate_grad_basis_dot_grad_field(
            element.basis_fields()
        ).coefficients,
        total,
    )



import fastfem.mesh as m

points = [
    m.Point(0, 0),
    m.Point(1, 0),
    m.Point(1, 1),
    m.Point(0.5, 1.5),
    m.Point(0, 1),
    m.Point(-0.5, 0.5),
    m.Point(0, 0),
]
lines = [m.Line(points[i], points[i + 1], number_of_nodes=10) for i in range(6)]
surface = m.Surface(lines, element_type="quadrangle")
mesh = m.mesh()



import numpy as np

import fastfem.fields.numpy_similes as fnp


def test_einsum_with_single_field(field_of_different_shape_orders_testsep):
    shapes = [(2, 3), (4, 1), (1, 2)]
    arr = np.random.rand(*shapes[0], *shapes[1], *shapes[2])
    factors = [np.random.rand(s[1]) for s in shapes]
    result = np.einsum("ijklmn,j,l,n", arr, *factors)
    for field in field_of_different_shape_orders_testsep(shapes, arr):
        tmp = fnp.einsum(field.shape_order_inverse[0], "ij,j", field, factors[0])
        tmp = fnp.einsum(field.shape_order_inverse[1], "ij,j", tmp, factors[1])
        tmp = fnp.einsum(field.shape_order_inverse[2], "ij,j", tmp, factors[2])
        np.testing.assert_allclose(tmp.coefficients, result)



import numpy as np

import fastfem.elements as elems
import fastfem.mesh as ffm
import fastfem.plotter as fplt
from fastfem.fields import ShapeComponent
from fastfem.fields import numpy_similes as fnp

# ========[ create a mesh element ]========
m = ffm.create_a_rectangle_mesh(
    horizontal_length=1,
    vertical_length=1,
    nodes_in_horizontal_direction=40,
    nodes_in_vertical_direction=40,
    element_type="triangle",
    file_name=None,
)
mesh = elems.LinearSimplexMesh2D(m)

# ========[   Initial Condition    ]========


def initial_condition(x, y):
    sigma = 0.03
    return np.exp(((x - 0.3) ** 2 + (y - 0.2) ** 2) / (-2 * sigma**2))


def initial_condition_dot(x, y):
    return np.zeros(x.shape)


X = mesh.position_field.point[0].coefficients
Y = mesh.position_field.point[1].coefficients

# wave field
U = mesh.Field(initial_condition(X, Y), is_const=False)
# wave field first derivative
Udot = mesh.Field(initial_condition_dot(X, Y), is_const=False)

# ========[     Run Parameters     ]========

c = 0.2  # wave speed
t = 0  # initial time

dt = 0.01  # time step size
tmax = 3  # sim end time

# ========[    Initialize Plot     ]========

visualizer = fplt.VisualMesh(m)
visualizer.plot_mesh()


# ========[       Time Loop       ]========

c2 = mesh.Field(c**2, is_const=True)
inv_mass_matrix = fnp.linalg.inv(
    fnp.moveaxis(
        mesh.mass_matrix(), (ShapeComponent.STACK, 0), (ShapeComponent.BASIS, 1)
    ),
    ShapeComponent.BASIS,
)


def solve_accel(wavefield):
    return -fnp.einsum(
        ShapeComponent.BASIS,
        "ij,j->i",
        inv_mass_matrix,
        mesh.integrate_grad_basis_dot_grad_field(wavefield, jacobian_scale=c2),
    )


Uddot = solve_accel(U)
n_steps = int(tmax / dt)
U_array = np.zeros((n_steps, U.coefficients.size))
for i in range(n_steps):
    # use Newmark-beta
    U += dt * Udot + (0.5 * dt**2) * Uddot
    U_array[i] = U.coefficients
    Udot += 0.5 * dt * Uddot
    Uddot = solve_accel(U)
    Udot += 0.5 * dt * Uddot
    t += dt

# ========[       Plotting       ]========

# plotting for a random index
visualizer.plot_data(U_array[150])

# animating
visualizer.animate_data(1 / dt, tmax, U_array, force=True)



import numpy as np
import pytest

from fastfem.elements.element2d import Element2D
from fastfem.elements.linear_simplex2d import LinearSimplex2D
from fastfem.elements.spectral_element2d import SpectralElement2D
from fastfem.fields import Field
from fastfem.fields import numpy_similes as fnp
from fastfem.fields.field import ShapeComponent

# ======================================================================================
#                                  Adding elements
#     When adding new elements, insert them into the dictionary `elements_to_test`. The
# key used to insert the element will be referenced in the name of the test
# parameterization. Additionally, to test the interpolate_... methods, a collection of
# points can be inserted into `element_test_localcoords`, which sould be an array of
# shape (...,2).
#
# This test suite will verify the basic functionality of the element,
# while things like integration will only be tested insofar as validating things such
# as field stacking. For an element to have a complete collection of tests, one should
# additionally have the following tested separately:
#
#   - interpolate_field(field,x,y). Agreement between different configurations of
#       stack_shape, point_shape or x,y shapes will be tested,
#       as well as linearity of solutions. So, the developer only needs to test a
#       spanning set of the scalar field space.
#
#   - compute_field_gradient(field, pos_matrix). As above, the developer only needs to
#       test a spanning set of the scalar field space. This test can be skipped by
#       appending the name of the element to
#       `elements_for_field_gradient_finitediff_test`, which tests
#       compute_field_gradient according to interpolate_field(),
#       reference_element_position_field() and finite differencing. It is assumed
#       that compute_field_gradient yields an exact value, and that nonlinear
#       transformations need not be checked.
#
#   - integrate_field(pos_matrix, field, jacobian_scale),
#   - integrate_basis_times_field(pos_matrix, field, indices, jacobian_scale)
#   - integrate_grad_basis_dot_field(pos_matrix,field,is_field_upper_index,indices,
#       jacobian_scale)
#   - integrate_grad_basis_dot_grad_field(pos_matrix,field,indices,jacobian_scale)
#       The same agreements will be verified (except with bilinearity of field and
#       jacobian_scale instead of linearity), so the developer only needs to test a
#       spanning set of the triple Cartesian product of the scalar field space and a
#       satisfactory set of transformed elements. Linear transformations on pos_matrix
#       will be tested by this suite, so only nonlinear transformations need to be
#       verified. Additionally, agreement with different index configurations will be
#       verified, so only indices=None needs to be taken as a case. Since
#       grad_basis_dot_field requires a vector field, so testing needs only be tested
#       with multiples of the elementary basis of R2.
#
#       Under most cases, integrate_grad_basis_dot_grad_field should be equivalent to
#       integrate_grad_basis_dot_field(..., compute_field_grad(field), ...). The test
#       for this is written in this test suite, and can be run by adding the element
#       name to `elements_for_grad_basis_dot_grad_field_test`
#
# ======================================================================================
elements_to_test = dict()
element_test_localcoords = dict()
elements_for_field_gradient_finitediff_test = set()
elements_for_grad_basis_dot_grad_field_test = set()
elements_for_ref_element_def_grad_is_identity_test = set()

# register linearsimplex2D
elements_to_test["linearsimplex"] = LinearSimplex2D()
element_test_localcoords["linearsimplex"] = np.array(
    [[0, 0], [1, 0], [0, 1], [0.5, 0.5], [0.25, 0.25]]
)
elements_for_field_gradient_finitediff_test.add("linearsimplex")
elements_for_grad_basis_dot_grad_field_test.add("linearsimplex")
elements_for_ref_element_def_grad_is_identity_test.add("linearsimplex")

# register spectral elements of different orders
for i in [3, 4, 5]:
    testname = f"spectral{i}"
    elements_to_test[testname] = SpectralElement2D(i)
    elements_for_field_gradient_finitediff_test.add(testname)
    elements_for_grad_basis_dot_grad_field_test.add(testname)
    elements_for_ref_element_def_grad_is_identity_test.add(testname)
    test_pts = np.empty((2 * i, 2 * i, 2))
    test_pts[:, :, 0] = np.linspace(-1, 1, 2 * i)[:, np.newaxis]
    test_pts[:, :, 1] = np.linspace(-1, 1, 2 * i)[np.newaxis, :]
    element_test_localcoords[testname] = test_pts

# ======================================================================================


@pytest.fixture(scope="module", params=elements_to_test.keys())
def element(request):
    return elements_to_test[request.param]


@pytest.fixture(scope="module", params=element_test_localcoords.keys())
def element_and_local_coords(request):
    key = request.param
    coords = element_test_localcoords[key]
    assert (
        key in elements_to_test
    ), f"Element {key} found in element_test_localcoords, but not elements_to_test."
    assert coords.shape[-1] == 2, (
        f"Last axis of element_test_localcoords['{key}'] is not size 2! shape is"
        f" {coords.shape}."
    )
    return elements_to_test[request.param], coords


def test_basis_and_reference_shapes(element: Element2D):
    """Validates the shapes of basis_fields() and reference_element_position_field()
    against basis_shape().
    """
    shape = element.basis_shape()
    basis = element.basis_fields()

    assert basis.basis_shape == shape, (
        "basis_fields() should have basis_shape == basis_shape(). "
        + f"basis_fields().basis_shape: {basis.basis_shape} basis_shape(): {shape}."
    )
    assert basis.point_shape == tuple(), (
        "basis_fields() should be a stack of scalar fields. Instead, point_shape =="
        f" {basis.point_shape}."
    )

    ref_elem_pts = element.reference_element_position_field()

    assert ref_elem_pts.basis_shape == shape, (
        "reference_element_position_field() should have basis_shape == basis_shape(). "
        + "reference_element_position_field().basis_shape:"
        f" {ref_elem_pts.basis_shape} basis_shape(): {shape}."
    )

    assert ref_elem_pts.stack_shape == tuple(), (
        "reference_element_position_field() should have stack_shape == tuple()."
        f" reference_element_position_field().stack_shape: {ref_elem_pts.stack_shape}"
    )

    assert ref_elem_pts.point_shape == (2,), (
        "reference_element_position_field() should have point_shape == (2,)."
        f" reference_element_position_field().point_shape: {ref_elem_pts.point_shape}"
    )


@pytest.mark.parametrize("elem", elements_for_ref_element_def_grad_is_identity_test)
def test_reference_deformation_gradient(elem: str):
    """Validates the reference position matrix's deformation gradient, which should
    be the identity.
    """
    element = elements_to_test[elem]
    def_grad = element.compute_field_gradient(
        element.reference_element_position_field()
    )
    np.testing.assert_allclose(def_grad.coefficients - np.eye(2), 0, atol=1e-6)


def random_basis_coefs(element: Element2D, num_trials: int | None = None):
    # maybe Hypothesis would have been a good idea?

    # default num_trials to twice the field space dimension
    basis_shape = element.basis_shape()
    trial_count = (
        num_trials if num_trials is not None else np.prod(basis_shape, dtype=int) * 2
    )

    for _ in range(trial_count):
        yield (np.random.rand(*basis_shape) * 2 - 1)


def test_interpolate_field_linearity(
    element_and_local_coords: tuple[Element2D, np.ndarray],
):
    element, coords = element_and_local_coords

    basis_shape = element.basis_shape()

    x = coords[..., 0]
    y = coords[..., 1]
    basis_field = element.basis_fields()
    interped_fields = element.interpolate_field(basis_field, x, y)
    collapse_str = "ABCDEFGHIJKLMNOPQURSTUVWXYZ"[: len(element.basis_shape())]
    for field_coefs in random_basis_coefs(element):
        expected = fnp.einsum(
            ShapeComponent.STACK,
            f"{collapse_str},{collapse_str}",
            interped_fields,
            field_coefs,
        ).coefficients
        np.testing.assert_allclose(
            element.interpolate_field(
                Field(basis_shape, tuple(), field_coefs), x, y
            ).coefficients,
            expected,
            atol=1e-8,
        )


@pytest.mark.parametrize("elem", elements_for_field_gradient_finitediff_test)
def test_compute_field_gradient(elem: str):
    assert elem in element_test_localcoords, (
        "The automatic test_compute_field_gradient test requires a sample of local"
        f" coordinates, but {elem} is not a key in element_test_localcoords."
    )
    element = elements_to_test[elem]
    coords = element_test_localcoords[elem]
    basis_field = element.basis_fields()

    # =======verify local derivs=======
    grads = element.compute_field_gradient(basis_field)
    x_padded = coords[..., 0]
    y_padded = coords[..., 1]

    grad_at_pts = element.interpolate_field(grads, x_padded, y_padded)

    # use central finite differencing
    eps = 1e-5
    grad_finite_diff_x = (
        element.interpolate_field(basis_field, x_padded + eps, y_padded)
        - element.interpolate_field(basis_field, x_padded - eps, y_padded)
    ).coefficients / (2 * eps)
    grad_finite_diff_y = (
        element.interpolate_field(basis_field, x_padded, y_padded + eps)
        - element.interpolate_field(basis_field, x_padded, y_padded - eps)
    ).coefficients / (2 * eps)
    np.testing.assert_allclose(
        grad_at_pts.coefficients
        - np.stack((grad_finite_diff_x, grad_finite_diff_y), axis=-1),
        0,
        atol=(eps**2) * 1e2,  # central diff has error O(h^2). give a constant factor
    )


@pytest.mark.parametrize("elem", elements_for_grad_basis_dot_grad_field_test)
def test_integrate_grad_basis_dot_grad_field(elem: str, affine_transforms):
    element = elements_to_test[elem]
    field = element.basis_fields()
    jac_scale = field.stack[..., *(np.newaxis for _ in field.basis_shape)]
    ref_points = element.reference_element_position_field().coefficients

    for name, T in affine_transforms:
        points = Field(element.basis_shape(), (2,), T(ref_points))
        grads = element.compute_field_gradient(field, points)
        stiffness = element.integrate_grad_basis_dot_field(
            points, grads, None, jac_scale
        )
        stiffness_computed = element.integrate_grad_basis_dot_grad_field(
            points, field, None, jac_scale
        )
        np.testing.assert_allclose(
            stiffness_computed.coefficients,
            stiffness.coefficients,
            err_msg=(
                "Failed agreement between integrate_grad_basis_dot_field and"
                f" integrate_grad_basis_dot_field for element transformed by '{name}'"
            ),
            verbose=True,
            atol=1e-8,
        )


def test_mass_matrix(element, affine_transforms):
    basis_field = element.basis_fields()
    basis_shape = element.basis_shape()
    ref_points = element.reference_element_position_field().coefficients
    basis_size = np.prod(basis_shape, dtype=int)

    for name, T in affine_transforms:
        pos_matrix = Field(element.basis_shape(), (2,), T(ref_points))

        mat_by_integ_basis = element.integrate_basis_times_field(
            pos_matrix, basis_field
        )
        mat_by_method = element.mass_matrix(pos_matrix)

        np.testing.assert_allclose(
            mat_by_method.coefficients, mat_by_integ_basis.coefficients, atol=1e-8
        )

        # and also test a few elements individually:
        for _ in range(10):
            indI = np.unravel_index(np.random.randint(basis_size), basis_shape)
            indJ = np.unravel_index(np.random.randint(basis_size), basis_shape)
            mat_inds = element.mass_matrix(pos_matrix, indices=indI + indJ)
            np.testing.assert_allclose(
                mat_inds.coefficients,
                mat_by_method.coefficients[..., *(indI + indJ)],
                atol=1e-8,
            )


def test_interpolate_field_gradient(
    element_and_local_coords: tuple[Element2D, np.ndarray], affine_transforms
):
    element, coords = element_and_local_coords
    basis_shape = element.basis_shape()

    x = coords[..., 0]
    y = coords[..., 1]
    x_padded = x[..., *(np.newaxis for ax in basis_shape)]
    y_padded = y[..., *(np.newaxis for ax in basis_shape)]
    basis_field = element.basis_fields()
    interped_grads = element.interpolate_field(
        element.compute_field_gradient(basis_field), x_padded, y_padded
    )
    np.testing.assert_allclose(
        element.interpolate_field_gradient(
            basis_field, x_padded, y_padded
        ).coefficients,
        interped_grads.coefficients,
        atol=1e-8,
    )
    ref_points = element.reference_element_position_field().coefficients
    for name, T in affine_transforms:
        pos_matrix = Field(element.basis_shape(), (2,), T(ref_points))

        interped_grads = element.interpolate_field(
            element.compute_field_gradient(basis_field, pos_matrix),
            x_padded,
            y_padded,
        )
        np.testing.assert_allclose(
            element.interpolate_field_gradient(
                basis_field, x_padded, y_padded, pos_matrix
            ).coefficients,
            interped_grads.coefficients,
            atol=1e-8,
        )


def test_incompatible_fields_throw(element: Element2D):
    pos_field = element.reference_element_position_field()
    pos_stack = pos_field.stack[np.newaxis].stack[np.zeros(10, dtype=int)]
    field_stack = Field(
        pos_field.basis_shape, tuple(), np.zeros((2,) + pos_field.basis_shape)
    )
    field_stack_incomp = Field(
        pos_field.basis_shape, tuple(), np.zeros((3,) + pos_field.basis_shape)
    )
    bad_basis = Field(
        pos_field.basis_shape + (2,), tuple(), np.zeros(pos_field.basis_shape + (2,))
    )
    bad_basis_posfield = Field(
        pos_field.basis_shape + (2,), (2,), np.zeros(pos_field.basis_shape + (2, 2))
    )
    X_incomp = np.array([0, 0, 0, 0])
    Y_incomp = np.array([0, 0, 0])

    with pytest.raises(ValueError):  # incompatible stacks
        element.interpolate_field(field_stack, X_incomp, Y_incomp)
    with pytest.raises(ValueError):  # wrong basis
        element.interpolate_field(bad_basis, 0, 0)
    with pytest.raises(ValueError):  # wrong basis (no pos_field)
        element.compute_field_gradient(bad_basis)
    with pytest.raises(ValueError):  # incompatible stacks
        element.compute_field_gradient(field_stack, pos_field=pos_stack)
    with pytest.raises(ValueError):  # posfield shape != (2,)
        element.compute_field_gradient(field_stack, pos_field=field_stack)
    with pytest.raises(ValueError):  # wrong basis
        element.compute_field_gradient(bad_basis, pos_field=pos_stack)
    with pytest.raises(ValueError):  # incompatible stacks (no posfield)
        element.interpolate_field_gradient(field_stack, X_incomp, Y_incomp)
    with pytest.raises(ValueError):  # wrong basis (no posfield)
        element.interpolate_field_gradient(bad_basis, 0, 0)
    with pytest.raises(ValueError):  # incompatible stacks
        element.interpolate_field_gradient(field_stack, 0, 0, pos_stack)
    with pytest.raises(ValueError):  # incompatible stacks from X,Y
        element.interpolate_field_gradient(field_stack, X_incomp, Y_incomp, pos_field)
    with pytest.raises(ValueError):  # wrong basis
        element.interpolate_field_gradient(
            Field(tuple(), tuple(), 1), 0, 0, bad_basis_posfield
        )
    with pytest.raises(ValueError):  # posfield shape != (2,)
        element.interpolate_field_gradient(field_stack, 0, 0, field_stack)
    for integfunc in [
        element.integrate_field,
        element.integrate_basis_times_field,
        element.integrate_grad_basis_dot_field,
        element.integrate_grad_basis_dot_grad_field,
    ]:
        with pytest.raises(ValueError):  # posfield shape != (2,)
            integfunc(field_stack, field_stack)
        with pytest.raises(ValueError):  # wrong basis
            integfunc(bad_basis_posfield, Field(tuple(), tuple(), 1))
        with pytest.raises(ValueError):  # jac incompatible
            integfunc(pos_field, field_stack, jacobian_scale=field_stack_incomp)
        with pytest.raises(ValueError):  # field incompatible
            integfunc(pos_field, field_stack_incomp, jacobian_scale=field_stack)
        with pytest.raises(ValueError):  # posfield incompatible
            integfunc(pos_stack, field_stack, jacobian_scale=field_stack)
        with pytest.raises(ValueError):  # posfield shape != (2,)
            integfunc(field_stack, field_stack, jacobian_scale=field_stack)
    with pytest.raises(ValueError):  # wrong basis
        element.mass_matrix(bad_basis_posfield)
    with pytest.raises(ValueError):  # posfield shape != (2,)
        element.mass_matrix(field_stack)
    with pytest.raises(ValueError):  # jac incompatible
        element.mass_matrix(pos_stack, jacobian_scale=field_stack_incomp)



import numpy as np
import pytest

from fastfem.fields.field import Field, ShapeComponent

_SHAPES_COMPARE_: list[tuple[tuple[int, ...], tuple[int, ...]]] = [
    (tuple(), tuple()),
    (tuple(), (0,)),
    (tuple(), (1,)),
    (tuple(), (2,)),
    (tuple(), (3, 2)),
    ((1,), (0,)),
    ((1,), (1,)),
    ((1,), (2,)),
    ((1,), (3, 2)),
    ((2,), (3,)),
    ((1, 2), (2, 1)),
    ((1, 1, 1, 5), (2, 2, 2, 5)),
    ((1, 1, 1, 5), (2, 2, 2, 4)),
    ((2, 2, 1, 1), (3,)),
]
_SHAPES_COMPARE_TRIPLES_: list[
    tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]
] = [
    *((a, b, tuple()) for a, b in _SHAPES_COMPARE_),
    ((1,), (2,), (3,)),
]


def are_shapes_compatible(*shapes: tuple[int, ...]):
    try:
        np.broadcast_shapes(*shapes)
    except ValueError:
        return False
    return True


_SHAPES_COMPARE: list[tuple[tuple[int, ...], tuple[int, ...], bool]] = [
    (a, b, are_shapes_compatible(a, b)) for a, b in _SHAPES_COMPARE_
]
_SHAPES_COMPARE_TRIPLES: list[
    tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...], bool]
] = [(a, b, c, are_shapes_compatible(a, b, c)) for a, b, c in _SHAPES_COMPARE_TRIPLES_]


# probably a better way of doing this,
# but we need to create a creator of generators, since one may call
# "for ... in itergen" multiple times.
class _itergen:
    def __init__(self, genfunc):
        self.genfunc = genfunc

    def __iter__(self):
        return self.genfunc()


@pytest.fixture
def comparison_two_shapes():
    def loop_shapes():
        for a, b, valid in _SHAPES_COMPARE:
            yield (a, b, valid)
            yield (b, a, valid)

    return _itergen(loop_shapes)


@pytest.fixture
def comparison_three_shapes():
    def loop_shapes():
        for a, b, c, valid in _SHAPES_COMPARE_TRIPLES:
            # yes, this isn't exhaustive of all permutations.
            yield (a, b, c, valid)
            yield (a, c, b, valid)
            yield (c, b, a, valid)

    return _itergen(loop_shapes)


def field_from_perm(shapes, coef, perm):
    return Field(
        shapes[perm[ShapeComponent.BASIS]],
        shapes[perm[ShapeComponent.POINT]],
        coef,
        perm,
    )


@pytest.fixture(
    params=[
        (0, 1, 2),
        (0, 2, 1),
        (1, 0, 2),
        (1, 2, 0),
        (2, 0, 1),
        (2, 1, 0),
    ]
)
def permutation(request):
    return request.param


@pytest.fixture
def field_of_different_shape_orders_testsep(permutation):
    def f(shapes, coef):
        yield field_from_perm(shapes, coef, permutation)

    return f


@pytest.fixture
def field_of_different_shape_orders():
    def f(shapes, coef):
        yield field_from_perm(shapes, coef, (0, 1, 2))
        yield field_from_perm(shapes, coef, (0, 2, 1))
        yield field_from_perm(shapes, coef, (1, 0, 2))
        yield field_from_perm(shapes, coef, (1, 2, 0))
        yield field_from_perm(shapes, coef, (2, 0, 1))
        yield field_from_perm(shapes, coef, (2, 1, 0))

    return f
